/*
	The Malete project - the Z39.2/Z39.50 database framework of OpenIsis.
	Version 0.9.x (patchlevel see file Version)
	Copyright (C) 2001-2004 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	see README for more information
EOH */

/*
	$Id: malete.c,v 1.2 2005/04/25 15:49:25 kripke Exp $
	malete main
*/

#include "tool.h"

static int usage ()
{
	eRr( 0, "malete "TOOL_VERSION" core "CORE_VERSION" pw "PW_VERSION
#ifndef BUILD_NOCDS
		" cds"
#endif
#ifndef BUILD_NOIIF
		" iif"
#endif
	);
	return 0;
}


static int qCb ( QLoop *self )
{
	Ptr p;
	unsigned i, vsz = self->qdx->vsz;
	const unsigned char *v = self->vals;
	char buf[512];
	int len = qRdKey(self->qdx, buf, sizeof buf, &self->cur);

	for ( i=self->nvals; i--; v += vsz) {
		qRdVal(&p, v, self->qdx->ptr);
		eOut( 0, "%.*s\t%u\t%d\t%u\t%u", len, buf,
			p.rid, p.tag, p.pos>>16, p.pos&0xffff);
	}
	return 0;
}	/* qCb */



/* ************************************************************
*/

int main ( int argc, const char **argv )
{
	List arg;
	Ses s;
	List rec;
	Fld  opt;
	Db  *db[2] = {0,0};
	char *p;
	int i;

	lInit(&rec, "test");
	CINIT(&arg);
	sInit(ses = &s);

	if ( LOG_VERBOSE >= env.log )
		SCPY(env.err, env.opt);

	for ( i=1; i<3 && i<LLEN(env.opt); i++ ) { /* open 1 or 2 dbs */
		if ( VPRI(&opt, env.opt->fld+i) ) {
			if ( !(db[i-1] = dOpen(p = VDUPZ( &opt ))) )
				return 1;
			mFree( p );
		}
	}

	if ( !VPRI(&opt, env.opt->fld) )
		return usage();

	switch ( opt.val[0] ) {
	case 'c':
#ifndef BUILD_NOCDS
		if ( VEQZ(&opt, "cdsimp") ) {
			if ( !db[0] ) return cdsuse();
			cdsimp(db[0], env.opt->fld);
		}
		else if ( VEQZ(&opt, "cdsexp") ) {
			if ( !db[0] ) return cdsuse();
			cdsexp(db[0], env.opt->fld);
		}
#endif
		break;
	case 'd':
		if ( VEQZ(&opt, "dbinfo") ) {
			if ( !db[0] ) goto dbs1;
			eRr(0, "db %s path %s has %d records",
				db[0]->nam, db[0]->pat, db[0]->rdx.mid);
			i = db[0]->rdx.typ;
			eRr(0, "mrx type 0x%02x: %d pos + %d len + %d fields = %d bytes",
				i, 4+(i>>4), 3+(i>>2&3), i&3, db[0]->rdx.ptl);
			i = db[0]->qdx.typ;
			eRr(0,
	"mqd type 0x%02x: blocksize %d, %d byte %s values, %d byte %scompressed keys",
				i, 1<<((i&QDX_SIZMSK)+QDX_LEAFSH),
				db[0]->qdx.vsz, QDX_LEAFPV&i ? "plain" : "pointer",
				db[0]->qdx.ksz, QDX_COMPRS&i ? "" : "un");
			i = db[0]->qdx.ptr;
			if ( !(QDX_LEAFPV&db[0]->qdx.typ) )
				eRr(0,
		"mqd pointers 0x%02x: %d record id + %d tag + %d pos bytes %s first",
					i, QDX_RIDMIN+((QDX_RIDMSK&i)>>QDX_RIDSH), (QDX_TAGMSK&i)>>QDX_TAGSH,
					QDX_POSMSK&i, QDX_FULTXT&i ? "tag" : "rid");
			if ( db[0]->opt )
				SCPYR(env.out, db[0]->opt);
		}
		break;
	case 'h':
		return usage();
	case 'i':
#ifndef BUILD_NOIIF
		if ( VEQZ(&opt, "iifimp") ) {
			if ( !db[0] ) return iifuse();
			iifimp(db[0], env.opt->fld);
		}
		else if ( VEQZ(&opt, "iifexp") ) {
			if ( !db[0] ) return iifuse();
			iifexp(db[0], env.opt->fld);
		}
#endif
		break;
	case 'q':
		if ( VEQZ(&opt, "qdump") ) {
			QLoop ql;
			if ( !db[0] ) goto dbs1;
			env.out->snk = fSinkl;
			eOut(0, ""); /* eats one */
			memset( &ql, 0, sizeof(ql) );
			ql.qcb = qCb;
			ql.qdx = &db[0]->qdx;
			qLoop( &ql );
		}
		else if ( VEQZ(&opt, "qload") ) {
			if ( !db[0] ) goto dbs1;
			qLoadf(&db[0]->qdx, &env.in);
		}
		else if ( VEQZ(&opt, "qchk") ) {
			QLoop ql;
			if ( !db[0] ) goto dbs1;
			memset( &ql, 0, sizeof(ql) );
			ql.qcb = 0;
			ql.qdx = &db[0]->qdx;
			qLoop( &ql );
		}
		else if ( VEQZ(&opt, "qset") ) {
			Key key;
			FIL_DEFBUF(env.in);
			if ( !db[0] ) goto dbs1;
			memset( &key, 0, sizeof(key) );
			key.val.len = db[0]->qdx.vsz;
			while ( fGets(&fb) && fb.l ) {
				if ( qMkKeyVal(&db[0]->qdx, &key, fb.p, fb.l) ) {
					eRr(LOG_VERBOSE, "bad key at line %d offset %d", fb.n, fb.o);
					continue;
				}
				if ( qSet(&db[0]->qdx, &key) )
					break;
			}
			eRr(LOG_INFO, "did %d lines %d bytes", fb.n, fb.o);
		}
		break;
	case 'r':
		if ( VEQZ(&opt, "rdump") ) {
			if ( !db[0] ) goto dbs1;
			for (i=1; i<=db[0]->rdx.mid; i++) {
				dRead(&env.out->lst, db[0], i);
				SEOR(env.out);
			}
		}
		else if ( VEQZ(&opt, "rload") ) {
			FIL_DEFBUF(env.in);
			while ( fGetr(&env.out->lst, &fb) ) {
				if ( db[0] )
					dWrite(db[0], env.out->lst.fld, 0);
				SEOR(env.out);
			}
		}
		break;
	case 's':
		if ( VEQZ(&opt, "server") )
			return server();
	}

	return 0;
dbs1:
	eRr(ERR_INVAL, "command %.*s needs a db", opt.len, opt.val);
	return 1;
}	/* main */
