/*
	The Malete project - the Z39.2/Z39.50 database framework of OpenIsis.
	Version 0.9.x (patchlevel see file Version)
	Copyright (C) 2001-2003 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	see README for more information
EOH */

/*
	$Id: pw.c,v 1.1 2005/01/24 18:01:20 kripke Exp $
	Malete patchwork functions
*/

#include "pw.h"


Ses *ses;


/* ************************************************************
	object
*/

int OPT_STDCALL mPipe ( Obj *self, Msg msg )
{
#if 1
	(void)self; (void)msg; return ERR_INVAL;
#else
	int rv = ERR_INVAL, eflg = env.flg;
	const char *p, *q, *e;
	Sink l1, l2, *pin, *pout, *eout = env.out;
	lInit( pout = &l1, 0 );
	lInit( pin = &l2, 0 );
	lCpy( pin, msg, 0 );
	LADD( pout, -1, msg->val, msg->len );
	env.out = pout;
	env.flg |= ENV_BUFFER;
	for ( p = msg->val, q = 0, e = msg->val + msg->len; e > p; ) {
		if ( ( TAB == *p++ && ( '|' == *p || msg->val == p-2 ) ) || e == p ) {
			if ( q ) {
				char *s = pin->fld->val;
				pin->fld->val += q - msg->val;
				pin->fld->len = ( p - q );
				if ( e > p )
					--pin->fld->len;
				if ( ( rv = self->snd( self, pin->fld ) ) )
					goto cleanup;
				pin->fld->val = s;
				pin->fld->len = msg->len;
				lReset( pin );
				env.out = pin;
				pin = pout;
				pout = env.out;
			}
			q = '|' == *p ? p+1 : p;
		}
	}
	LCPY( eout, pin );
cleanup:
	env.out = eout;
	env.flg = eflg;
	lClr( pin );
	lClr( pout );
	return rv;
#endif
}	/* mPipe */


int OPT_STDCALL mComp ( Obj *self, Msg msg )
{
	const char *e = msg->val + msg->len;
	char *q, *p;
	int rv = ERR_INVAL;
	for ( q = 0, p = msg->val, e = msg->val + msg->len; e > p; ) {
		if ( ( TAB == *p++ && ( ';' == *p || msg->val == p-2 ) ) || e == p ) {
			if ( q ) {
				msg->val = q;
				msg->len = (p - q);
				if ( e > p )
					--msg->len;
				if ( 0 > (rv = self->snd(self, msg)) )
					return rv;
			}
			q = ';' == *p ? p+1 : p;
		}
	}
	return rv;
}	/* mComp */


int OPT_STDCALL mComm ( Obj *self, Msg msg )
{
	EADDF(msg);
	if ( -1 <= msg->tag )
		return 0;
	if ( msg[1].tag > msg->tag+1 ) /* not properly embedded */
		msg[1].tag = msg->tag+1;
	return self->snd(self, msg+1);
}	/* mComm */


int OPT_STDCALL oObj ( Obj *self, Msg msg )
{
	eRr(LOG_VERBOSE, "oObj '%.*s'", msg->len, msg->val);
	if ( msg->len )
		switch (*msg->val) {
		case '|': return mPipe(self, msg);
		case ';': return mComp(self, msg);
		case '#': return mComm(self, msg);
		case '=': return 0;
		case '.': return dispatch(msg);
		}
	else
		return eOut(0, "#\t0");
	return eRr(ERR_INVAL, "unrecognized message '%.*s'", msg->len, msg->val); 
}	/* oObj */



/* ************************************************************
	struct
*/
int OPT_STDCALL oStruct ( Struct *self, Msg msg )
{
	HEntry *he;
	int nlen;
	char *p = msg->val, *e = p + msg->len;
	if ( !msg->len || CT_A != lat1ct[(unsigned char)*p] )
		return oObj((Obj*)self, msg);
	eRr(LOG_VERBOSE, "oStruct '%.*s'", msg->len, msg->val);
	/* must address child */
	for ( ;p < e; p++ )
		switch (*p) {
		case TAB:
			break;
		case '.':
			goto dot;
		}
	return eRr(ERR_INVAL, "bad struct msg '%.*s'", p-msg->val, msg->val);
dot:
	nlen = p++ - msg->val;
	if ( !self->reg || !(he = hGet(self->reg, msg->val, nlen)) ) {
		Obj *cld;
		if ( !self->fac || !(cld = self->fac(self, msg->val, nlen)) )
			return ERR_INVAL;
		if ( !self->reg ) {
			self->reg = (HTable *) mAlloc( sizeof(HTable) );
			hInit(self->reg, "reg");
		}
		he = hSet(self->reg, msg->val, nlen, cld);
	}
	if ( he && he->obj ) { 
		msg->len -= nlen + 1;
		msg->val = p; 
		return  ((Obj *) he->obj)->snd( (Obj *) he->obj, msg ); 
	} 
	return eRr(ERR_INVAL, "no such child '%.*s'", nlen, msg->val);
} /* oStruct */


int OPT_STDCALL oSes ( Ses *self, Msg msg )
{
	self->req++;
	if ( msg->len && '.' == *msg->val ) { /* rooted */
		msg->val++;
		msg->len--;
	} else if (self->at) { /* prepend default */
		/* TODO */
	}
	return oStruct((Struct*)self, msg);
}	/* oSes */


Obj* OPT_STDCALL facMain ( Ses *self, const char *key, int len )
{
	char *name = mDupz(key, len);
	Obj *ret;

	eRr(LOG_VERBOSE, "facMain '%.*s'", len, key);
	(void)self; /* could be used for authorization etc. */
	/* we should scan env.opt for defined objects like remote servers
		however, db is the default anyway
	*/
	ret = (Obj*)newDbo(name);
	mFree(name);
	return ret;
}	/* facMain */


void sInit (Ses *self)
{
	memset(self, 0, sizeof(Ses));
	self->snd = (disp*)oSes;
	self->fac = (factory*)facMain;
	self->out = env.out;
	self->err = env.err;
	/* TODO: check env opts for option settings */
	self->r = 20;
	self->q = 20;
	self->s = 10000;
}	/* sInit */
