/*
	The Malete project - the Z39.2/Z39.50 database framework of OpenIsis.
	Version 0.9.x (patchlevel see file Version)
	Copyright (C) 2001-2003 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	see README for more information
EOH */
#ifndef PW_H

/*
	$Id: pw.h,v 1.2 2005/02/23 17:09:52 kripke Exp $
	interface of the Malete patchwork
*/

#define PW_VERSION "1.0"


#include "core.h"


typedef struct Obj Obj;


/* ****************************************************************** */
/*                                                                    */
/* HASH TABLE                                                         */
/*                                                                    */
/* ****************************************************************** */

#define HEDEFKEYLEN 1

typedef struct HEntry HEntry;
/* hmmm ... maybe this should become part of Obj */
struct HEntry {
	HEntry *nxt;
	Obj    *obj;
	int     hv; /* hash value of this entry */
#ifndef NDEBUG
	int     dpt;
#endif
	int     len;
	char    key[HEDEFKEYLEN]; /* not 0 terminated */
};

/* hmmm ... maybe this should become part of Struct */
typedef struct HTable {
	List *list;
	int   mask;
	int   nume; /* # of entries */
	int   dsize; /* # bytes of unlinked entries */
#ifndef NDEBUG
	int   coll; /* # of collisions */
#endif
} HTable;

#define HSIZ(t) (-((t)->list->fld->tag)-1) /* number of buckets */ 
#define HSHIFT 1 /* rebuild factor: HSIZ(t) << HSHIFT */ 
#define HESIZ(he)  (sizeof(HEntry)+(he)->len-HEDEFKEYLEN) /* entry size */ 
#define HINITIAL 32 /* initial # of buckets */
#define HMAXLOAD 75 /* max load factor */
#define HCURLOAD(t) ( (100*(t)->nume) /  HSIZ(t) ) /* current load */ 
#define HSHRINKLOAD 25 /* shrink, if HCURLOAD(t) < HSHRINKLOAD */
#define HMAXFRAG 50 /* max fragmentation factor */ 
#define HCURFRAG(t) ( (100*((t)->dsize)) / LSIZ((t)->list) ) /* current frag */


extern HEntry* OPT_STDCALL hGet ( HTable *table, const char *key, int len );
extern HEntry* OPT_STDCALL hAdd ( HTable *table, const char *key, int len );
extern HEntry* OPT_STDCALL hSet ( HTable *t, const char *key, int len, Obj *o );
extern int OPT_STDCALL hDel ( HTable *table, const char *key, int len );
extern HTable* OPT_STDCALL hInit ( HTable *table, const char* header );
extern void OPT_STDCALL hFini ( HTable *table );

/* free al resources held by table */
#define HFREE( table ) do { \
		hFini( (table) ); \
		mFree( table ); \
	} while (0)

#ifndef NDEBUG
/* debug statistics */
extern void OPT_STDCALL hstat ( HTable *table );
#	define HSTAT(t) hstat(t);
#else
# define HSTAT(t)
#endif



/* ****************************************************************** */
/*                                                                    */
/* OBJECTS AND MESSAGES                                               */
/*                                                                    */
/* ****************************************************************** */

typedef Fld * Msg;
typedef struct Qry Qry; /* query implemented by dbo */


/*
	dispatch message "msg", starting at object obj.

	-----------------------------------
	Syntax and semantics of messages

	A message is basically an array of fields - usually of length 1. The leader
	of this array containes the messagename, denoting a method, and possibly
	options. The mesagename has to be given in the leader's primary value, the
	options in following subfields.  Data to be processed is contained in the
	following fields - if any. The messagename may contain subobject specifiers,
	which have to be separated by a '.'. Besides this the evaluation of the
	given leader is solely in the responsibility of the method invoked. The
	message evaluation may change the given message. The significant length of a
	messagename may be limited to four bytes. A leader should not be longer
	than DEFBLKLEN bytes, though some messages may allow longer sizes. Some
	messagenames may use special characters denoting a special treatment of the
	given leader - thus such characters should not be used for ordinary
	messagenames. The result of the evaluation, if any, is always written to
	env.out.
*/
typedef int OPT_STDCALL disp ( Obj *self, Msg msg );

/*
	the general form of an object
	send it a message using that->snd(msg)
	In general, snd must return <0 on error, >=0 else
*/
struct Obj {
	disp *snd; /* method implementation */
};
/**
	base implementation for Obj handling pipes and such
*/
extern int OPT_STDCALL oObj ( Obj *self, Msg msg );


typedef struct Struct Struct;

typedef Obj* OPT_STDCALL factory ( Struct *self, const char *key, int len );
/*
	an object with variable number of childs
*/
struct Struct {
	disp    *snd; /* method implementation */
	HTable  *reg; /* child registry */
	factory *fac; /* child factory */
};
/**
	base implementation for Struct handling child dispatching
*/
extern int OPT_STDCALL oStruct ( Struct *self, Msg msg );


/**
	session -- the root object
*/
typedef struct Ses {
	disp    *snd;
	HTable  *reg; /* main registry */
	factory *fac; /* main factory */
	Sink    *out; /* main env.out for this session */
	Sink    *err; /* env.err for this session */
	unsigned rid; /* last rid written */
	Qry     *qry; /* array of q queries */
	lolo     req; /* request counter */
	/* options */
	char    *at; /* default target */
	unsigned r; /* max # of records per read, 0 = unlimited */
	unsigned q; /* max # of open query cursors != 0 */
	unsigned s; /* max # of results in search set, 0 = unlimited */
} Ses;
extern int  OPT_STDCALL oSes ( Ses *self, Msg msg );

/*
	default main factory (i.e. for the session).
	scans the environment options for defined objects
*/
extern Obj* OPT_STDCALL facMain ( Ses *self, const char *key, int len );

/* init session */
extern void sInit (Ses *s);


extern Ses *ses; /* current session */

/*
	main dispatcher
*/
#define dispatch( msg ) ses->snd((Obj*)ses, msg)


/*
	object linked to a session on some remote server
*/
typedef struct Srv {
	disp *snd; /* method implementation */
	file *con; /* connection */
} Srv;
/**
	interface a stream
*/
extern int OPT_STDCALL oSrv ( Srv *self, Msg msg );


/*
	database object
	NOTE that these are childs of session, so all status is per session.
*/
typedef struct Dbo {
	disp    *snd;
	HTable  *reg; /* views etc */
	factory *fac; /* child factory */
	Db      *db;  /* shared by the Dbos */
	Qry     *qry; /* chain of queries */
	unsigned qid; /* number of last query, 0 = none */
	unsigned rid; /* number of last written rec, 0 = none */
	/* options */
	unsigned p; /* position limit for snapshot read, 0 = none */
} Dbo;
/**
	handle db access
*/
extern int OPT_STDCALL oDbo ( Dbo *self, Msg msg );

extern Dbo *newDbo ( char *name ); /* constructor */
extern void delDbo ( Dbo *self ); /* destructor */


/* ****************************************************************** */
/*                                                                    */
/* HANDLERS FOR VARIOUS STANDARD MESSAGES                             */
/*                                                                    */
/* ****************************************************************** */


/*
	Process a pipe message '|'.
	The leader is parsed for the subfields '|'. Each of which should contain a
	valid messagename, options may be given in following subfields terminated by
	the next '|' subfield or the end of the leader. The output of the first
	method is used as data input for the next and so on. The first submessage
	may start directly after the primary field w/o the '|' identifier.
*/
extern int OPT_STDCALL mPipe ( Obj *self, Msg msg );

/*
	Process a compound message ';'.
	The leader is parsed for the subfields ';'. Each of which should contain a
	valid messagename, options may be given in following subfields terminated by
	the next ';' subfield or the end of the leader. Each message is processed
	separately, given data is used for all messages. The first submessage may
	start directly after the primary field w/o the ';' identifier.
*/
extern int OPT_STDCALL mComp ( Obj *self, Msg msg );

/*
	Process a comment message '#'.
*/
extern int OPT_STDCALL mComm ( Obj *self, Msg msg );


#define PW_H
#endif /* PW_H */
