/*
	The Malete project - the Z39.2/Z39.50 database framework of OpenIsis.
	Version 0.9.x (patchlevel see file Version)
	Copyright (C) 2001-2004 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	see README for more information
EOH */

/*
	$Id: srv.c,v 1.1 2005/01/24 18:04:11 kripke Exp $
	malete server
*/
#ifndef WIN32
#include <sys/types.h>
#include <sys/time.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <unistd.h>
#include <stdlib.h> /* exit */
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <string.h> /* memcpy et al */
#include <signal.h>
#include <errno.h>
#include <sys/wait.h>
#else
#define WIN32_LEAN_AND_MEAN
#define NONAMELESSUNION
/* #include <winsock2.h> -lws2_32 */
#include <winsock.h>	/* -lwsock32 "good" enough ? */
#define close closesocket
#endif

#include "tool.h"


static void stdsrv ( file f )
{
	List in;
	FIL_DEFBUF(f);
	lInit(&in, 0);
	while ( fGetr(lClr(&in), &fb) ) {
		dispatch(in.fld);
		SEOR(env.out);
	}
}	/* srv */


static char buf[0x2000];
static int  fill;
static int  crlf;
static int  wrotehead;
static int  broken;
static unsigned sock; /* response socket */
enum {
	SOL, /* at start of line */
	TAG, /* in tag (only digits seen in line) */
	VAL  /* in val (some non-digit seen) */
};

static int flushsock ()
{
	int got = send(sock, buf, fill, 0);
	if ( 0 < got ) {
		if ( !(fill -= got) )
			return 0;
		memmove(buf, buf+got, fill);
		return 0;
	}
	fill = 0;
	return broken = eRr(ERR_NO, "could not send");
}


static void sinksock (Sink *lo, int eor)
{
	Fld *f = lo->lst.fld, *e = f + RLEN(f);

	if ( (unsigned)-1 == sock )
		goto reset;
	if ( wrotehead )
		f++;
	for ( ; f < e; f++ ) {
		int maxfill = sizeof(buf)-13-f->len, l, n;
		char *p, *v;
		if ( 0 > maxfill )
			maxfill = 0;
		/* make sure we have enough room */
		while ( fill > maxfill )
			if ( flushsock() )
				goto reset;
		p = buf+fill;
		if ( wrotehead ) {
			p += i2a(p, f->tag);
			*p++ = TAB;
		} else {
			if ( 10>b36val[(unsigned)*f->val] ) {
				*p++ = 'W';
				*p++ = TAB;
			}
			wrotehead = 1;
		}
		v = f->val;
		l = f->len;
		/* write field value */
		while ( l >= (n = buf+sizeof(buf) - p) ) {
			memcpy(p, v, n);
			fill = sizeof(buf);
			if ( flushsock() )
				goto reset;
			p = buf+fill;
		}
		memcpy(p, v, l);
		p += l;
		*p++ = LF;
		fill = p - buf;
	} /* for */
	if ( eor ) { /* add a \n and flush */
		if ( sizeof(buf) == fill && flushsock() ) /* pathological */
			goto reset;
		buf[fill++] = LF;
		while ( fill && !flushsock() )
			;
	}
reset:
	lReset(&lo->lst);
}	/* sinksock */


/*
	serve one request on sock
*/
static int srv1 ()
{
#define CR 13
	int state = SOL, neg = 0;
	char *p, *e, *n;
	unsigned l;
	List in;
	Fld *f = 0;

	crlf = 0;
	lInit(&in, 0);

wantmore: /* really a for, but we need multi-level continue anyway */
	if ( 0 >= (fill = recv(sock, p = buf, sizeof(buf), 0)) )
		return eRr(LOG_WARN, "got %d in recv", fill);
	LOG_DBG(LOG_DEBUG, "got %d bytes", fill);
	e = p + fill;
havemore:
	switch (state) {
	case SOL:
		switch (*p) {
		case CR:
			if (e == p+1) goto wantmore; /* ignore */
			if (LF != p[1]) break; /* weird crap */
			p++;
		case LF: /* got blank line */
			if ( e > ++p )
				return eRr(ERR_INVAL, "got %d excess bytes", e-p);
			goto gotit; /* break 2 */
		}
		LNEWF(&in, 0, 80);
		if ( !f && (CT_IS(D,*p) || '-' == *p) ) {
			eRr(LOG_WARN, "no message name");
			LNEWF(&in, 0, 80);
		}
		f = LLAST(&in);
		state = TAG;
		if ( (neg = '-' == *p) && e == ++p ) goto wantmore;
	case TAG:
		while ( CT_IS(D,*p) ) {
			f->tag = f->tag*10 + b36val[(unsigned char)*p];
			if ( e == ++p ) goto wantmore; /* continue 2 */
		}
		if ( neg )
			f->tag = -f->tag;
		LOG_DBG(LOG_DEBUG, "got tag %d", f->tag);
		state = VAL;
		if ( TAB == *p && e == ++p ) goto wantmore;
	case VAL:
		n = memchr(p, LF, e-p);
		if ( !n )
			l = e-p;
		else if ( (l = n-p) && 13 == n[-1] ) {
			if ( crlf )
				l--;
			else if ( f == in.fld ) {
				l--;
				crlf = 1;
				LOG_DBG(LOG_DEBUG, "detected crlf");
			}
		}
		LAPP(&in, p, l);
		if ( n ) {
			state = SOL;
			if ( e > (p = n+1) )
				goto havemore;
		}
		goto wantmore;
	}
gotit:
	fill = 0;
	wrotehead = 0;
	broken = 0;
	env.out->off = 0;
	lClr(&env.out->lst);
	dispatch(in.fld);
	SEOR(env.out);
	return broken;
}	/* srv1 */


/* ************************************************************
*/

int server ()
{
	Ses s;
	Fld opt = { 0, 0, 0 };
	fd_set lst; /* listening socks */
	fd_set all;
	unsigned fdlen = 0;
#ifndef WIN32
	unsigned kids = 0;
#endif

	sInit(ses = &s);
	FD_ZERO(&lst);
	FD_ZERO(&all);
	if ( env.opt ) for ( opt.val = 0; vGet(&opt, env.opt->fld, "S"); )
		switch (opt.tag) {
			static const int yes = !0;
			struct sockaddr_in si;
#ifndef WIN32
			struct sockaddr_un su;
#endif
			struct sockaddr *sa;
			int sal;
			Fld info;
		case 'S':	/* socket */
			info = opt; /* for message */
#ifndef WIN32
			if ( opt.len && '/'==*opt.val ) { /* unix socket */
				if ( opt.len > sizeof(su.sun_path)-1 ) {
					eRr(ERR_INVAL, "unix socket name '%.*s' too long", opt.len, opt.val);
					return 1;
				}
				memset(&su, 0, sizeof(su));
				memcpy(su.sun_path, opt.val, opt.len);
				su.sun_path[opt.len] = 0;
				su.sun_family = AF_UNIX;
				unlink(su.sun_path);
				sa = (struct sockaddr*)&su;
				sal = sizeof(su);
				sock = socket(PF_UNIX, SOCK_STREAM, 0);
			} else
#else
			{
				WSADATA wsadata;
				memset(&wsadata, 0, sizeof(wsadata));
				if ( WSAStartup(2, &wsadata) ) {
				/* funny enough, winsock 1 says ok :) */
					eRr(ERR_INVAL, "could not get winsock");
					return 1;
				}
			}
#endif
			{ /* TCP socket */
				memset(&si, 0, sizeof(si));
				si.sin_family = AF_INET;
				si.sin_addr.s_addr = INADDR_ANY;
				si.sin_port = htons(2042);
				if ( opt.len ) { /* standard */
					Fld port = opt;
					char *end = memchr(opt.val, ':', opt.len);
					if ( end /* host:port */
						|| CT_D!=lat1ct[(unsigned char)*opt.val] /* hostname */
						|| memchr(opt.val, '.', opt.len)
					) {
						if ( !end ) end = opt.val+opt.len;
						if ( 9 >= (end - opt.val)
							&& !memcmp("localhost",opt.val,end - opt.val)
						)
							si.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
						else { /* parse dotted decimal */
						}
						if ( (port.len = opt.len - (end - opt.val)) ) {
							port.len--;
							port.val = end+1;
						}
					}
					if ( port.len )
						si.sin_port = htons(V2I(&port));
				} else {
					info.val = "*:2042";
					info.len = 6;
				}
				sa = (struct sockaddr *)&si;
				sal = sizeof(si);
				sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP/*6*/);
			}
			if ( setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes)) ) {
				eRr(ERR_NO, "setsockopt");
				return 1;
			}
			if ( bind(sock, sa, sal) ) {
				eRr(ERR_NO, "could not bind to '%.*s'", info.len, info.val);
				return 1;
			}
			if ( listen(sock, 64) ) {
				eRr(ERR_NO, "listen failed");
				return 1;
			}
			eRr(LOG_INFO, "listening on '%.*s'", info.len, info.val);
			FD_SET(sock, &lst);
			FD_SET(sock, &all);
			if ( fdlen < sock+1 )
				fdlen = sock+1;
		}
	/* done with options */
	if ( !fdlen ) {
		stdsrv(env.in);
		return 0;
	}
	env.out->snk = sinksock;
	for (;;) {
		fd_set sele = all;
		int got = select(fdlen, &sele, 0, 0, 0);
		if ( 0 >= got ) {
			eRr(ERR_NO, "select said %d", got);
			continue;
		}
		for ( sock=0; got--; ) {
			while (!FD_ISSET(sock, &sele))
				sock++;
			if ( FD_ISSET(sock, &lst) ) {
				struct sockaddr peer;
				unsigned /*socklen_t is broken*/ plen = sizeof(peer);
				unsigned nsock = accept(sock, &peer, &plen);
#ifndef WIN32
				if ( ENV_EXCL != env.wri ) { /* shared or read only */
					if ( fork() ) {
						close(nsock);
						kids++;
						continue;
					}
					sock = nsock;
					while ( !srv1() )
						;
					exit(0);
				}
#endif
				FD_SET(nsock, &all);
				if ( fdlen < nsock+1 )
					fdlen = nsock+1;
				continue;
			}
			/* else serve a request on this socket */
			/* TODO: use the socket's session */
			if ( srv1() ) {
				close(sock);
				FD_CLR(sock, &all);
			}
		}
#ifndef WIN32
		for ( ;kids && 0 < waitpid(-1, 0, WNOHANG); kids--)
			;
#endif
	}
	return 0;
}	/* server */
