/*
	The Malete project - the Z39.2/Z39.50 database framework of OpenIsis.
	Version 0.9.x (patchlevel see file Version)
	Copyright (C) 2001-2003 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	see README for more information
EOH */

/*
	$Id: sys.c,v 1.2 2005/04/25 15:49:25 kripke Exp $
	I/O support for the openisis library.
*/
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <limits.h> /* PATH_MAX */
#include <sys/types.h>

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#define NONAMELESSUNION
#	include <windows.h>
#	include <sys/timeb.h>
#else	/* have operating system */
#	define __USE_UNIX98
/*
#	define _POSIX_MAPPED_FILES
# define _POSIX_SYNCHRONIZED_IO
*/
#	include <fcntl.h> /* or try unixio */
#	ifndef O_SYNC
#		if defined( __FreeBSD__ )
#			define O_SYNC O_FSYNC
#		else
#			define O_SYNC 0
#		endif
#	endif
#	include <sys/stat.h>
#	include <sys/file.h> /* flock */
#if defined(__linux__) && !defined(__USE_GNU)
#	define __USE_GNU /* weird trigger to declare MREMAP_MAYMOVE */
#endif
#	include <sys/mman.h>
#	include <sys/socket.h>
#	include <unistd.h>
#endif /* WIN32 */

#include <sys/time.h> /* gettimeofday */
#include <time.h> /* localtime */


#include "core.h"



/* ************************************************************
	private data
*/


/* ************************************************************
	private functions
*/

static void *memordie ( int size )
{
	void *p = malloc( size );
	if ( p )
		return p;
	exit( 71/*EX_OSERR*/ ); /* exit handlers MUST NOT alloc memory */
	return 0;
}

/* ************************************************************
	private functions
*/


/* ************************************************************
	public data
*/
Env env;



/* ************************************************************
	public functions
*/


void *mAlloc ( int size )
{
	void *p = memordie( size );
	memset(p, 0, size);
	return p;
}	/* mAlloc */


void  mFree ( void *mem )
{
	if ( mem )
		free( mem );
}	/* mFree */


void *mDup ( const void *str, int sz )
{
	void *m = memordie( 0<=sz ? sz : (sz = strlen( str ) + 1) );
	memcpy( m, str, sz );
	return m;
}	/* mDup */


char *mDupz ( const char *str, int sz )
{
	char *m = memordie( sz+1 );
	memcpy( m, str, sz );
	m[sz] = 0;
	return m;
}	/* mDupz */


LBlk *mBlkAlloc ( int size )
{
	LBlk *b = memordie( (((char*)(&((LBlk*)0)->byt))-(char*)0) + size );
	b->nxt = 0;
	b->siz = size;
	memset(b->byt, 0, size);
	return b;
}	/* mBlkAlloc */


List *mListAlloc ( const char *name )
{
	return lInit(memordie(sizeof(List)), name);
}	/* mListAlloc */


int tUpd ( lolo *tm )
{
	lolo o;
#ifdef WIN32
	struct _timeb tb;
	_ftime( &tb );
	if ( !tm )
		return tb.time;
	o = *tm;
	*tm = tb.time*LOLO(1000) + tb.millitm;
#else
	struct timeval tv;
	gettimeofday( &tv, 0 );
	if ( !tm )
		return tv.tv_sec;
	o = *tm;
	*tm = tv.tv_sec*LOLO(1000) + tv.tv_usec/1000;
#endif
	return (int)(*tm - o);
}	/* tUpd */


static int tLoc ( struct tm *t, lolo *tm )
{
	lolo x;
	time_t tt;
	if ( !tm || !*tm )
		tUpd( tm ? tm : (tm = &x) );
	tt = (time_t)(*tm / 1000);
#ifdef WIN32 /* did I mention it's not threadsafe ? */
	*t = *localtime( &tt );
#else
	localtime_r( &tt, t );
#endif
	return (int)(*tm % 1000);
}	/* tLoc */


/* logically belongs to uti.c; here so tGtf may consider inlining */
void u2a0 ( char *p, unsigned u, unsigned n )
{
	for ( ; n--; u/=10 )
		p[n] = '0' + (u % 10);
}	/* u2a0 */


int tGtf ( char *buf, lolo *tm )
{
	struct tm t;
	int millis = tLoc( &t, tm );

	u2a0(buf, 1900+t.tm_year, 1900+t.tm_year);
	u2a0(buf+4, 1+t.tm_mon, 2);
	u2a0(buf+6, t.tm_mday, 2);
	u2a0(buf+8, t.tm_hour, 2);
	u2a0(buf+10, t.tm_min, 2);
	u2a0(buf+12, t.tm_sec, 2);
	buf[14] = 0;
	return millis;
}	/* tGtf */


char *tGtfm ( char *buf, lolo *tm )
{
	u2a0(buf+14, tGtf(buf, tm), 3);
	buf[18] = 0;
	return buf;
}	/* tGtfm */


#if 0 /* unused */
void tSleep ( lolo tm )
{
#ifdef WIN32
	Sleep( tm );
#else
	struct timespec ts;
	ts.tv_sec = tm / 1000;
	ts.tv_nsec = 1000000 * (int)(tm % 1000);
	nanosleep( &ts, 0 );
#endif
}	/* tSleep */
#endif


#ifndef WIN32
static const char *lmode (int f)
{
	switch ( (FIL_BLOCK|FIL_TLOCK|FIL_WR) & f ) {
	case FIL_BLOCK|FIL_WR: return "wait ex";
	case FIL_BLOCK: return "wait sh";
	case FIL_TLOCK|FIL_WR: return "test ex";
	case FIL_TLOCK: return "test sh";
	}
	return "unlock";
}
#endif


int fOpen ( file *fil, const char *name, int flags )
{
	int faillvl = FIL_TRY&flags ? LOG_VERBOSE : ERR_NO;

#ifndef WIN32
	int fd = -1;
	int f = !(FIL_WR&flags) ? O_RDONLY : ((FIL_RD&flags) ? O_RDWR : O_WRONLY);

	*fil = FIL_NONE;
	if (!name) { /* tempfile */
		const char tpl[] = "/tmp/maleteXXXXXX";
		char nam[sizeof tpl];
		memcpy(nam, tpl, sizeof tpl);
		fd = mkstemp(nam);
		if (0 > fd)
			return eRr(ERR_NO, "could not create tmpfile");
		eRr(LOG_VERBOSE, "tempfile '%s'", nam);
		unlink(nam);
		*fil = fd;
		return 0;
	}
	if ( FIL_CREAT & flags && FIL_WR & flags ) f |= O_CREAT;
	if ( FIL_SYNC & flags ) f |= O_SYNC;
	if ( FIL_TRUNC & flags ) f |= O_TRUNC;
	fd = open( name, f, 00664 ); /* let umask finetune */
	if ( 0 > fd )
		return eRr( faillvl, "could not open '%s' %x", name, flags );
	if ( FIL_FLOCK & flags ) {
		struct flock fl;
		memset(&fl, 0, sizeof(fl));
		fl.l_type   = FIL_WR&flags ? F_WRLCK : F_RDLCK;
		fl.l_whence = SEEK_SET;
		fl.l_start  = 0;
		fl.l_len    = 0;
		eRr(LOG_VERBOSE, "%s perm lock on '%s'", lmode(flags), name);
		if ( fcntl(fd, FIL_BLOCK&flags ? F_SETLKW : F_SETLK, &fl) ) {
			close(fd);
			return eRr(ERR_NO, "err %s perm lock on '%s'", lmode(flags), name);
		}
	}
	*fil = fd;
#else
	HANDLE h;
	int acc = 0;
	int shr = FILE_SHARE_READ | (FIL_FLOCK&flags ? 0 : FILE_SHARE_WRITE);
	int cre = (FIL_CREAT & flags && FIL_WR & flags)
		? (FIL_TRUNC & flags) ? CREATE_ALWAYS : OPEN_ALWAYS
		: (FIL_TRUNC & flags) ? TRUNCATE_EXISTING : OPEN_EXISTING;
	int f = (FIL_WR&flags) ? FILE_ATTRIBUTE_ARCHIVE : FILE_ATTRIBUTE_NORMAL;
	char nam[128];

	*fil = FIL_NONE;
	if (!name) { /* tempfile */
		GetTempPath(sizeof nam, nam);
		GetTempFileName(nam, "mtp", 0, nam);
		name = nam;
		eRr(LOG_VERBOSE, "tempfile '%s'", nam);
		acc = GENERIC_READ|GENERIC_WRITE;
		shr = 0;
		cre = TRUNCATE_EXISTING;
		f = FILE_ATTRIBUTE_TEMPORARY|FILE_FLAG_DELETE_ON_CLOSE;
	} else {
		if ( FIL_RD & flags ) acc |= GENERIC_READ; /* FILE_READ_DATA; */
		if ( FIL_WR & flags ) acc |= GENERIC_WRITE;
		if ( FIL_SYNC & flags ) f |= FILE_FLAG_WRITE_THROUGH;
	}
	f |= FILE_FLAG_RANDOM_ACCESS;

	h = CreateFile( name, acc, shr, 0, cre, f, 0 );
	if ( INVALID_HANDLE_VALUE == h )
		return eRr( faillvl, "could not open '%s' %x", name, flags );
	*fil = h;
#endif
	return 0;
}	/* fOpen */


int fClose ( file *f )
{
	if ( FIL_NONE != *f ) {
#ifndef WIN32
		for ( fsync( *f ); close( *f ); errno = 0 ) 
			if ( EINTR != errno ) {
				eRr( LOG_SYSERR, "could not close file %d", *f );
				return errno ? -errno : -1;
			}
#else
		CloseHandle( (HANDLE)*f );
#endif
		*f = FIL_NONE;
	}
	return 0;
}	/* fClose */


int fSize ( file f )
{
#ifndef WIN32
	{
	struct stat s;
	return fstat( f, &s ) ? 0 : s.st_size;
	}
#else
	return GetFileSize( (HANDLE)f, 0 );
#endif
}	/* fSize */


unsigned fTime ( file f )
{
#ifndef WIN32
	struct stat s;
	return fstat( f, &s ) ? 0 : (unsigned)s.st_mtime;
#else
	FILETIME foo; /* time since 160101011200 UTC in hundred nanoseconds !!! */
	lulu bar;
	if ( ! GetFileTime( (HANDLE)f, 0, 0, &foo ) )
		return 0;
	bar = ((lulu)foo.dwHighDateTime<<32 | (lulu)foo.dwLowDateTime) / LULU(10000000);
	return (unsigned)(bar - LULU(11644473600));
#endif
}	/* fTime */


int fRead ( file *f, void *buf, unsigned count )
{
#ifndef WIN32
	int got = read( *f, buf, count );
	return 0 <= got ? got
		: eRr( ERR_NO, "could not read %d bytes from %d", count, *f );
#else
	DWORD got = 0;
	int ok = ReadFile( (HANDLE)*f, buf, count, &got, 0 );
	return ok ? got : eRr( ERR_NO, "could not read %d bytes", count );
#endif
}	/* fRead */


int fWrite ( file *f, const void *buf, unsigned count )
{
	{
#ifndef WIN32
	int got = write( *f, buf, count );
	return 0 <= got ? got
		: eRr( ERR_NO, "could not write %d bytes to %d", count, *f );
#else
	DWORD got = 0;
	int ok = WriteFile( (HANDLE)*f, buf, count, &got, 0 );
	return ok ? got
		: eRr( ERR_NO, "could not write %d bytes", count );
#endif
	}
}	/* fWrite */



int fSeek ( file *f, unsigned offset )
{
	{
#ifndef WIN32
	int got = (int)lseek( *f, offset, SEEK_SET );
	return (int)offset == got ? 0
		: eRr( ERR_NO, "could not seek to %d", offset );
#else
	return 0xffffffff /* "INVALID_SET_FILE_POINTER" */
		== SetFilePointer( (HANDLE)*f, offset, 0, FILE_BEGIN )
		? -1 : 0;
#endif
	}
}	/* fSeek */


int fPread ( file *f, void *buf, unsigned count, unsigned offset )
{
	{
#ifndef WIN32
	int got = pread( *f, buf, count, offset );
	return 0 <= got ? got
		: eRr( ERR_NO, "could not read %d bytes from %d at %d",
				count, *f, offset );
#else
	return 0xffffffff /* "INVALID_SET_FILE_POINTER" */
		== SetFilePointer( (HANDLE)*f, offset, 0, FILE_BEGIN )
		? -1 : fRead( f, buf, count );
#endif
	}
}	/* fPread */


int fPwrite ( file *f, const void *buf, unsigned count, unsigned offset )
{
	{
#ifndef WIN32
	int got = pwrite( *f, buf, count, offset );
	return 0 <= got ? got
		: eRr( ERR_NO, "could not write %d bytes to %d at %d",
				count, *f, offset );
#else
	return 0xffffffff /* "INVALID_SET_FILE_POINTER" */
		== SetFilePointer( (HANDLE)*f, offset, 0, FILE_BEGIN )
		? -1 : fWrite( f, buf, count );
#endif
	}
}	/* fPwrite */


int fTrunc ( file *f, unsigned length )
{
#ifndef WIN32
	ftruncate( *f, length );
	if ( length && (int)length != fSize(*f) ) {
		/* Solaris and FreeBSD guarantee that ftruncate can extend files,
			but Linux does not.
		*/
		static char nosh = 0;
		fPwrite(f, &nosh, 1, length-1);
	}
#else
	/* SetFileValidData( (HANDLE)*f, length ); XP only */
	return 0xffffffff /* "INVALID_SET_FILE_POINTER" */
		== SetFilePointer( (HANDLE)*f, length, 0, FILE_BEGIN )
		? -1 : SetEndOfFile( (HANDLE)*f )
		? 0 : eRr( ERR_NO, "could not truncate %d", length );
#endif
	return 0;
}	/* fTrunc */


#ifdef BUILD_SHMODE
int fLock ( file f, unsigned n, int flg )
{
	struct flock fl;

	memset( &fl, 0, sizeof(fl) );
	fl.l_type   = FIL_WR&flg ? F_WRLCK : flg ? F_RDLCK : F_UNLCK;
	fl.l_whence = SEEK_SET;
	fl.l_start  = n;
	fl.l_len    = 1;
	eRr( LOG_VERBOSE, "%s tmp lock %d", lmode(flg), n );
	return fcntl(f, FIL_BLOCK&flg ? F_SETLKW : F_SETLK, &fl)
		? eRr(ERR_NO, "err %s tmp lock %d", lmode(flg), n) : 0;
}	/* fLock */
#endif


int fMOpen ( FMap *fm, const char *name, int flags )
{
	int ret;
	fm->map = 0;
	fm->npg = 0;
	if ( (ret = fOpen(&fm->fil, name, flags)) )
		fm->flg = 0;
	else {
#ifdef WIN32 /* save mapping name */
		static const char pfx[] = "Malete";
		int l = strlen(name);
		char *c;
		fm->nam = mAlloc(sizeof(pfx)+l);
		memcpy(fm->nam, pfx, sizeof(pfx)-1);
		memcpy(c = fm->nam+sizeof(pfx)-1, name, l+1);
		for (; *c; c++) if ( '\\' == *c ) *c = '/'; /* must not contain \ */
#endif
		fm->flg = flags;
	}
	return ret;
}	/* fMOpen */


int fMClose ( FMap *fm )
{
	if ( fm->map )
		fMap(fm, 0);
#ifdef WIN32 /* save mapping name */
	if ( fm->nam ) {
		mFree(fm->nam);
		fm->nam = 0;
	}
#endif
	return fClose(&fm->fil);
}	/* fMClose */


int fMap ( FMap *fm, unsigned npg )
{
	if ( npg > fm->lim && fm->npg == (npg = fm->lim) )
		return npg;
#ifndef WIN32
	if ((FIL_WR & fm->flg) && npg && (((int)npg<<env.psh) > fSize(fm->fil)))
		fTrunc(&fm->fil, npg<<env.psh);
	if ( fm->map ) { /* unmap or remap */
#ifdef __linux__
		if ( npg ) {
			LOG_DBG(LOG_DEBUG, "remapping %d pages at %8x", fm->npg, fm->map);
			fm->map = mremap(fm->map, fm->npg<<env.psh, npg<<env.psh, MREMAP_MAYMOVE);
			fm->npg = 0;
			goto done;
		}
#endif
		munmap(fm->map, fm->npg<<env.psh);
	}
	fm->map = 0;
	fm->npg = 0;
	if (!npg)
		return 0;
	/* get new map */
	fm->map = mmap( 0, npg<<env.psh,
		PROT_READ | ((FIL_WR&fm->flg) ? PROT_WRITE : 0), MAP_SHARED, fm->fil, 0 );
#ifdef __linux__
done:
#endif
	LOG_DBG(LOG_DEBUG, "mapped %d pages at %8x", npg, fm->map);
	if ( MAP_FAILED != fm->map )
		return (int)(fm->npg = npg);
	fm->map = 0;
	return eRr(ERR_NO, "mmap failed");
#else
	/* awfully complicated here
		-- need  CreateFileMapping, MapViewOfFile
		stupid piece of shrott
		true mapping supported in NT family only, 9x copies to swap
	*/
	if ( fm->map ) {
		UnmapViewOfFile(fm->map);
		CloseHandle(fm->hdl);
	}
	fm->map = 0;
	fm->npg = 0;
	fm->hdl = 0;
	if (!npg)
		return 0;
	/*
	note that shared writing of a mapping does not work,
	because to be consistent, it needs to be the same mapping object,
	which we can not grow.
	However, windoze doesn't support shared writing anyway.
	*/
	{
	int size = fSize(fm->fil);
	if (FIL_WR&fm->flg || size > npg<<env.psh)
		size = npg<<env.psh;
	else
		npg = (size + env.psz-1)>>env.psh;
	if ( (fm->hdl = CreateFileMapping(fm->fil,0,
		(FIL_WR&fm->flg)?PAGE_READWRITE:PAGE_READONLY,0,size,fm->nam)) ) {
		if ( (fm->map = MapViewOfFile(fm->hdl,
			(FIL_WR&fm->flg)?FILE_MAP_WRITE:FILE_MAP_READ,0,0,size))
		)
			return fm->npg = npg;
		CloseHandle(fm->hdl);
		fm->hdl = 0;
		return eRr(ERR_NO, "MapViewOfFile %d failed", npg);
	}
	}
	return eRr(ERR_NO, "CreateFileMapping failed");
#endif
}	/* fMap */


int fMSync ( FMap *fm, unsigned page )
{
	char * buf = fm->map + (page<<env.psh);
	LOG_DBG(LOG_DEBUG, "msync %d %8x", page, buf);
#ifndef WIN32
	return msync( buf, env.psz, (
#ifdef BUILD_SHMODE
		ENV_SHARED == env.wri ? MS_INVALIDATE :
#endif
		0) | (FIL_SYNC&fm->flg ? MS_SYNC : MS_ASYNC) )
		? eRr( ERR_NO, "msync" ) : 0;
#else
	return FlushViewOfFile(buf, env.psz) ? 0 : eRr(ERR_NO, "msync");
#endif
}	/* fMSync */


int fSlurp ( char **buf, int sz, const char *name, int opt )
{
	int size;
	int ret;
	char *p = *buf;
	file f;

	if ( (ret = fOpen( &f, name, FIL_RD|(opt ? FIL_TRY : 0) )) ) 
		return ret;
	size = fSize( f );
	if ( 0 >= size )
		return size;
	if ( size > sz ) {
		eRr(ERR_INVAL, "file '%.*s' too big: %d > %d", 30, name, size, sz);
		goto done;
	}
	if ( ! p && !(p = mAlloc(size)) ) {
		ret = ERR_NOMEM;
		goto done;
	}
	ret = fRead( &f, p, size );
	if ( size == ret )
		*buf = p;
	else {
		if ( ret >= 0 )
			ret = eRr( ERR_IO, "OOPS! got %d of %d bytes from '%.*s'",
				ret, size, 30, name );
		if ( ! *buf )
			free( p );
	}
done:
	fClose( &f );
	return ret;
}	/* fSlurp */


int fGets ( FBuf *b )
{
	char *c;
	int n;
	if ( b->l ) {
		b->p += b->l;
		b->o += b->l;
		b->l = 0;
	}
	if ( b->m ) {
		b->p++;
		b->o++;
		b->m--;
		b->n++;
	}
	for (;;) { /* now have b->m bytes at b->p+b->l */
		if ( b->m ) { /* add all or part of this b->l */
			if ( (c = memchr(b->p+b->l, LF, b->m)) ) {
				b->m -= c - b->p - b->l;
				b->l = c - b->p;
				return !0;
			}
			b->l += b->m;
			b->m = 0;
			if ( b->l == b->s ) /* buffer exhausted */
				return !0;
			if ( b->p != b->b )
				memmove(b->b, b->p, b->l);
		}
		b->p = b->b;
		if ( 0 >= (n = fRead(&b->f, b->p+b->l, b->s-b->l)) ) /* eof */
			return b->l; /* but had line */
		b->m = n;
	}
}	/* fGets */


int fGetr ( List *l, FBuf *fb )
{
	if ( !fGets(fb) )
		return 0;
	while ( fb->l ) {
		int tag;
		int n = a2il(fb->p, fb->l, &tag);
		if ( !l->fld->tag ) { /* init rec */
			if ( n ) /* this is a data field, add blank header */
				LPREF(l, -1, 0);
			else /* use this one as header */
				tag = -1;
		}
		if ( n < (int)fb->l && TAB == fb->p[n] )
			n++;
		LADD(l, tag, fb->p+n, fb->l-n);
		if ( !fGets(fb) )
			break;
	}
	return !0;
}	/* fGetr */



static void fSinkf (Sink *lo, int eor, int full)
{
	Fld *f = lo->lst.fld, *e = f + RLEN(f);
	file to = (file)lo->dst;

	if ( FIL_NONE == to ) goto reset;
	for ( f++; f < e; f++ ) {
		char buf[8192];
		int l;
		if ( full ) {
			l = i2a(buf, f->tag);
			buf[l++] = '\t';
		} else {
			if ( !f->len )
				continue;
			l = 0;
		}
		if ( f->len < sizeof(buf)-l ) {
			memcpy(buf+l, f->val, f->len);
			buf[l += f->len] = '\n';
			fWrite(&to, buf, l+1);
		} else {
			if ( l )
				fWrite(&to, buf, l);
			if ( f->len )
				fWrite(&to, f->val, f->len);
			buf[0] = '\n';
			fWrite(&to, buf, 1);
		}
	}
	if ( eor )
		fWrite(&to, "\n", 1);
reset:
	lReset(&lo->lst);
}	/* fSinkf */

void fSinkl (Sink *lo, int eor)
{
	fSinkf(lo, eor, 0);
}	/* fSinkl */

void fSinkr (Sink *lo, int eor)
{
	Fld *f = lo->lst.fld;
	file to = (file)lo->dst;
	char buf[8192];
	int l;

	if ( FIL_NONE == to ) goto reset;
	if ( f->tag ) { /* not empty */
		if ( !lo->off && f->len ) { /* skip header if off or empty */
			if ( 10>b36val[(unsigned)*f->val] ) {/* starts with digit */
				buf[0] = 'W';
				buf[1] = TAB;
				l = 2;
			} else
				l = 0;
			if ( f->len < sizeof(buf)-l ) {
				memcpy(buf+l, f->val, f->len);
				buf[l += f->len] = '\n';
				fWrite(&to, buf, l+1);
			} else {
				if ( l )
					fWrite(&to, buf, l);
				if ( f->len )
					fWrite(&to, f->val, f->len);
				fWrite(&to, "\n", 1);
			}
		}
	}
	fSinkf(lo, eor, 1);
	if ( ! eor ) {
		lo->off += f->tag; /* no of fields passed by */
		return;
	}
reset:
	lo->off = 0;
	lClr(&lo->lst);
}	/* fSinkr */


int eOut ( int tag, const char *fmt, ... )
{
	if ( env.out ) {
		va_list ap;
		va_start( ap, fmt );
		lOut(&env.out->lst, tag, 0, fmt, ap);
		va_end( ap );
		SINK(env.out);
	}
	return tag;
}	/* eOut */


int eRr ( int tag, const char *fmt, ... )
{
	va_list ap;

	if ( ERR_NO == tag )
#ifndef WIN32
		switch ( errno ) {
		case EINTR: case EAGAIN:
			tag = ERR_AGAIN; break;
		case EFAULT: case ENAMETOOLONG: case ELOOP:
			tag = ERR_FAULT; break; /* structurally bad address */
		case EBADF: case ENOENT: case ENOTDIR:
			tag = ERR_BADF; break; /* file "does not exist" */
		case EIO: case EEXIST: case EISDIR: case EACCES: case ENXIO:
		case ENODEV: case EROFS: case ETXTBSY: case ENOSPC:
		case EPIPE: case ESPIPE:
			tag = ERR_IO; break; /* file "is not accessible" */
		case ENOMEM: case EMFILE: case ENFILE:
			tag = ERR_NOMEM; break;
		case EBUSY:
			tag = ERR_BUSY; break;
		default:
			tag = ERR_INVAL;
		}
#else
		tag = ERR_INVAL;
#endif
	if ( env.log > tag )
		return tag;

	if ( !env.err )
		return tag;
	va_start( ap, fmt );
	lOut(&env.err->lst, -tag, 0, fmt, ap);
	va_end( ap );

	/* logging afterburner */
	if ( ERR_NOMEM >= tag && tag >= ERR_AGAIN ) {
		char *syserr =
#ifndef WIN32
			errno ? strerror(errno) : 0;
#else
			0;
		char buf[256] = "";
		if ( 
			FormatMessage(
				FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS,
				0, GetLastError(), 0, buf, sizeof(buf)-1, 0 )
		)
			syserr = buf;
		buf[ sizeof(buf)-1 ] = 0;
#endif
		if ( syserr ) {
			LAPP(&env.err->lst, ": ", 2);
			LAPPS(&env.err->lst, syserr);
		}
	}
	SINK(env.err); /* just trigger the sink, buffering should be handled there */
	return tag;
}	/* eRr */



/*
	NOTE: no shared lib support
*/
void cFini ()
{
	dCloseAll();
	/* close env's buffers */
	if ( env.out ) env.out->snk(env.out, 1);
	if ( env.err ) env.err->snk(env.err, 1);
}	/* cFini */


void cInit ( List *opts, Sink *out, Sink *err )
{
	Sink **oe;
	Fld opt = { 0, 0, 0 };
	int i;

	if ( env.out )
		return;
	memset(&env, 0, sizeof(env));
	env.opt = opts;
#ifndef WIN32
	env.in = 0;
#else
	if ( INVALID_HANDLE_VALUE == (env.in = GetStdHandle(STD_INPUT_HANDLE)) )
		env.in = FIL_NONE;
#endif
	env.out = out;
	env.err = err;
	oe = (Sink**)&env.out;
	for ( i=0; i<2; i++ ) {
		oe = i ? &env.err : &env.out;
		if ( *oe )
			continue;
		*oe = (Sink*)mAlloc(sizeof(Sink));
#ifndef WIN32
		(*oe)->dst = (void*)(i + 1);
#else
		if ( INVALID_HANDLE_VALUE ==
			((*oe)->dst = GetStdHandle(i?STD_ERROR_HANDLE:STD_OUTPUT_HANDLE))
		)
			fOpen((file*)&(*oe)->dst, i ? "err.txt" : "out.txt", FIL_WR|FIL_CREAT);
#endif
		if ( i ) {
			lInit(&env.err->lst, "err");
			env.err->snk = fSinkl;
		} else {
			lInit(&env.out->lst, 0);
			env.out->snk = fSinkr;
		}
	}

	/* check pagesize */
#ifndef WIN32
	env.psz = sysconf(_SC_PAGESIZE); /* getpagesize(); */
#else
	{
		SYSTEM_INFO si;
		GetSystemInfo(&si);
		env.psz = si.dwPageSize; /* usually also 4k */
	}
#endif
	{
		int t = env.psz;
		env.psh = 0;
		while (t >>= 1) env.psh++;
		if ( env.psz != (unsigned)1<<env.psh || env.psh > CPU_PAGE_SHIFT ) {
			eRr(ERR_TRASH, "weirdness! pagesize %d shift %d max %d",
				env.psz, env.psh, CPU_PAGE_SHIFT);
			exit(1);
		}
		if ( env.psh < 12 ) { /* but using multiples is ok */
			eRr(LOG_WARN, "page_shift is %d !!!", env.psh);
			env.psz = 1<<(env.psh = 12);
		}
	}
	/* defaults */
	env.log = LOG_WARN;
	env.wri = ENV_EXCL;
	env.qml = env.rml = 1<<(30-env.psh); /* map 1 GB at most */

	if ( opts ) for ( opt.val = 0; vGet(&opt, opts->fld, "vw"); )
		switch (opt.tag) {
		case 'v':
			/*
			logging:
			first char selects level, a 'b' as second sets buffering
			*/
			if ( opt.len ) {
				switch ( *opt.val ) {
				case '0': case 'o': env.log = 0; break;
				case '1': case 'f': env.log = LOG_FATAL; break;
				case '2': case 's': env.log = LOG_IOERR /*LOG_SYSERR*/; break;
				case '3': case 'e': env.log = LOG_ERROR; break;
				case '4': case 'w': env.log = LOG_WARN; break;
				case '5': case 'i': env.log = LOG_INFO; break;
				case '6': case 'v': env.log = LOG_VERBOSE; break;
				case '7': case 'd': env.log = LOG_DEBUG; break;
				case '8': case 't': env.log = LOG_TRACE; break;
				case '9': case 'a': env.log = LOG_ALL; break;
				}
				if ( 1 < opt.len && 'b' == opt.val[1] )
					env.flg |= ENV_BUFFER;
			}
			break;
		case 'w': /* write/concurrency mode */
			if ( opt.len )
				switch ( *opt.val ) {
				case '0': env.wri = ENV_RO; break;
				case 's':
#ifdef BUILD_SHMODE
					env.wri = ENV_SHARED;
					env.flg |= ENV_MSYNC;
#else
					eRr(ERR_INVAL, "no shared mode build");
#endif
					break;
			}
		} /* if (in) for switch */

	atexit( cFini );
}	/* cInit */
