/*
	The Malete project - the Z39.2/Z39.50 database framework of OpenIsis.
	Version 0.9.x (patchlevel see file Version)
	Copyright (C) 2001-2003 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	see README for more information
EOH */

/*
	$Id: uti.c,v 1.1 2005/01/24 18:04:54 kripke Exp $
	utilities
*/

#include <string.h> /* strlen */
#include <stdarg.h>


#include "core.h"

/* ************************************************************
	data
*/
const char b36dig[36] = {
	'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f','g','h',
	'i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z'
};

const char b36val[256] = {
	36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,
	36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,
	36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,
	 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,36,36,36,36,36,36,
	36,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,
	25,26,27,28,29,30,31,32,33,34,35,36,36,36,36,36,
	36,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,
	25,26,27,28,29,30,31,32,33,34,35,36,36,36,36,36,
	36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,
	36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,
	36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,
	36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,
	36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,
	36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,
	36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,
	36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36
};

const unsigned char lat1up[256] = {
	  0,' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',
	' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',
	' ','!','"','#','$','%','&','\'','(',')','*','+',',','-','.','/',
	'0','1','2','3','4','5','6','7','8','9',':',';','<','=','>','?',
	'@','A','B','C','D','E','F','G','H','I','J','K','L','M','N','O',
	'P','Q','R','S','T','U','V','W','X','Y','Z','[','\\',']','^','_',
	'`','A','B','C','D','E','F','G','H','I','J','K','L','M','N','O',
	'P','Q','R','S','T','U','V','W','X','Y','Z','{','|','}','~',' ',
	' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',
	' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',
	' ',161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
	176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,
	'A','A','A','A','A','A','A','C','E','E','E','E','I','I','I','I',
	'T','N','O','O','O','O','O',215,'O','U','U','U','U','Y','T','S',
	'A','A','A','A','A','A','A','C','E','E','E','E','I','I','I','I',
	'T','N','O','O','O','O','O',247,'O','U','U','U','U','Y','T','Y'
};

const unsigned char lat1ct[256] = {
/* NUL  STX  SOT  ETX  EOT  ENQ  ACK  BEL */
  CT_W,CT_W,CT_W,CT_W,CT_W,CT_W,CT_W,CT_W,
/*  BS   HT   LF   VT   FF   CR   SO   SI */
  CT_W,CT_W,CT_W,CT_W,CT_W,CT_W,CT_W,CT_W,
/* DLE  DC1  DC2  DC3  DC4  NAK  SYN  ETB */
  CT_W,CT_W,CT_W,CT_W,CT_W,CT_W,CT_W,CT_W,
/* CAN   EM  SUB  ESC   FS   GS   RS   US */
  CT_W,CT_W,CT_W,CT_W,CT_W,CT_W,CT_W,CT_W,
/*blank  !    "    #    $    %    &    '  */
  CT_W,CT_S,CT_S,CT_S,CT_S,CT_S,CT_S,CT_S,
/*  (    )    *    +    ,    -    .    /  */
  CT_S,CT_S,CT_S,CT_S,CT_S,CT_S,CT_S,CT_S,
/*  0    1    2    3    4    5    6    7  */
  CT_D,CT_D,CT_D,CT_D,CT_D,CT_D,CT_D,CT_D,
/*  8    9    :    ;    <    =    >    ?  */
  CT_D,CT_D,CT_S,CT_S,CT_S,CT_S,CT_S,CT_S,
/*  @    A    B    C    D    E    F    G  */
  CT_S,CT_A,CT_A,CT_A,CT_A,CT_A,CT_A,CT_A,
  CT_A,CT_A,CT_A,CT_A,CT_A,CT_A,CT_A,CT_A, /* H-O */
  CT_A,CT_A,CT_A,CT_A,CT_A,CT_A,CT_A,CT_A, /* P-W */
/*  X    Y    Z    [    \    ]    ^    _  */
  CT_A,CT_A,CT_A,CT_S,CT_S,CT_S,CT_S,CT_I,
/*  `    a    b    c    d    e    f    g  */
  CT_S,CT_A,CT_A,CT_A,CT_A,CT_A,CT_A,CT_A,
  CT_A,CT_A,CT_A,CT_A,CT_A,CT_A,CT_A,CT_A,
  CT_A,CT_A,CT_A,CT_A,CT_A,CT_A,CT_A,CT_A,
/*  x    y    z    {    |    }    ~   DEL */
  CT_A,CT_A,CT_A,CT_S,CT_S,CT_S,CT_S,CT_S,
/* 32 C1 controls */
    0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
    0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
    0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
    0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
/* 32 mostly symbols */
    0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
    0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
    0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
    0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
/* 64 Latin alphas including 2 symbols */
  CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,
  CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,
  CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,  0 ,
  CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,
  CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,
  CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,
  CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,  0 ,
  CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,CT_L,CT_L
};

/* ************************************************************
	general utilities
*/


int a2il ( const char *p, int l, int *res )
{
	const char *s = p;
	const char *e;
	char op = 0;
	int d, ret = 0;
	if ( 0 > l ) { /* 0 terminated */
		while ( ' ' >= (unsigned char)*p )
			if ( !*p++ )
				return p - s -1;
		e = p+11; /* max useful, we bail out on 0 byte */
	} else
		for ( e = p+l; p < e && ' ' >= (unsigned char)*p; ) /* skip white */
			p++;
	if ( p < e )
		switch ( *p ) {
		case '-':
			if ( p+1 == e ) ret = 1; /* so sole '-' is -1 */
		case '~': /* sole ~ is ~0 */
		case '+':
			op = *p++;
		}
	if ( p < e && '0' == *p && ++p < e && 'x' == *p )
		while ( ++p < e && 16 > (d = b36val[(unsigned char)*p]) )
			ret = (ret<<4) + d;
	else
		for ( ;p < e && 10 > (d = b36val[(unsigned char)*p]); p++ )
			ret = 10*ret + d;
	if (0 != res)
		switch (op) {
		case '-': *res = -ret; break;
		case '~': *res = ~ret; break;
		default: *res = ret;
		}
	return p - s;
}	/* a2il */

int a2i ( const char *p, int l ) {
	int res;
	a2il (p, l, &res);
	return res;
}

int a2id ( const char *p, int l, int dflt ) {
	int res;
	if (0 > l) {
		l = (int) strlen (p);
	}
	if (l != a2il (p, l, &res)) {
		return dflt;
	}
	return res;
}


int i2a ( char *p, int i )
{
	if ( i ) {
		int min, dig = 0, j;
		if ( 0 < i ) 
			min = 0;
		else {
			min = 1;
			*p++ = '-';
			i = -i;
		}
		for ( j=i; j; j/=10 )
			dig++;
		for ( *(p+=dig)=0; i; i/=10 )
			*--p = '0' + (i % 10);
		return min+dig;
	}
	*p = '0';
	return 1;
}	/* i2a */


int u2a ( char *p, unsigned u )
{
	if ( u ) {
		char *d = p+10;
		for ( ; u; u/=10 )
			*--d = '0' + (u % 10);
		if ( p < d ) {
			char *e = p+10, *q = p;
			while ( d < e )
				*q++ = *d++;
			return q-p;
		}
		return 10;
	}
	*p = '0';
	return 1;
}	/* u2a */




/* ************************************************************
	field functions
*/

int vGet ( Fld *dst, const Fld *src, const char *opt )
{
	char *p, *e = src->val + src->len;
	const char *q;

	if ( !dst->val || (p = dst->val+dst->len) < src->val || dst->val > e ) {
		/* initial */
		p = src->val;
		if ( (!opt || '*' == *opt) && (!src->len || TAB == *p) ) {
			dst->tag = 0;
			dst->val = p; /* catch empty primary */
			return 1 + (dst->len = 0);
		}
	} else if ( p >= e )
		goto notfound;
	if ( p == src->val && (!opt || '*' == *opt) && TAB != *p )
		dst->tag = 0;
	else {
		for (;;) {
			if ( e == p )
				goto notfound;
			if ( TAB != *p++ )
				continue;
			if ( e == p ) /* bad trailing tab */
				goto notfound;
			/* p is on valid subfield indicator */
			if ( !opt ) /* take any */
				break;
			for (q=opt; *q; )
				if (*p == *q++)
					goto takeit; /* break 2 */
		}
	takeit:
		dst->tag = *p++;
	}
	dst->val = p;
	while ( e > p && TAB != *p )
		p++;
	return 1 + (dst->len = p - dst->val);
notfound:
	dst->val = 0;
	dst->len = 0;
	return 0;
}	/* vGet */


int vCmp ( const Fld *a, const Fld *b )
{
	unsigned l = a->len;
	int cmp;

	if ( l > b->len ) l = b->len;
	return (l && (cmp = memcmp(a->val, b->val, l)))
		? cmp : (int)a->len - (int)b->len;
}


int vGt ( const Fld *a, const Fld *b )
{
	unsigned l = a->len;
	int cmp = 0;

	if ( l > b->len ) l = b->len;
	return (l && 0 < (cmp = memcmp(a->val, b->val, l))) ? 1
		: !cmp && a->len > b->len;
}



/* ************************************************************
	record functions
*/

unsigned rSiz ( const Fld *r )
{
	unsigned ret = 0;
	const Fld *e = r + RLEN(r);

	while ( r < e )
		ret += (r++)->len;
	return ret;
}	/* rSiz */


const Fld *rGet ( const Fld *r, int tag, int *pos )
{
	const Fld *e = r + RLEN(r);
	const Fld *p = r + (pos ? *pos : 0);

	while ( ++p < e )
		if ( tag == p->tag ) {
			if ( pos )
				*pos = p-r;
			return p;
		}
	if ( pos )
		*pos = e-r;
	return 0;
}	/* rGet */


const Fld *rKey ( const Fld *r, int tag, const char *key )
{
	const Fld *e = r + RLEN(r);
	const Fld *d = 0;
	unsigned kl = strlen(key);

	while ( ++r < e )
		if ( tag == r->tag ) {
			if ( VKEY(r, key, kl) )
				return r;
			if ( !r->len || TAB == *r->val )
				d = r;
		}
	return d;
}	/* rKey */


const Fld *rDup ( const Fld *src, unsigned siz )
{
	int len = RLEN(src);
	const Fld *e = src + len;
	Fld *dst, *f;
	char *p;

	if ( !siz )
		siz = rSiz( src );
	f = dst = mAlloc( len*sizeof(Fld) + siz );
	p = ((char*)dst) + len*sizeof(Fld);
	for ( ;src < e; src++, f++ ) {
		f->tag = src->tag;
		memcpy( f->val = p, src->val, f->len = src->len );
		p += f->len;
	}
	return dst;
}	/* rDup */


int rSer ( char *buf, const Fld *f )
{
#if 0
	static const char repl[4] = { 0, VT, TAB, ' ' };
	const char *s, *e;
#endif
	const Fld *fe = REND(f);
	char *p = buf;

	while ( ++f < fe ) {
		p += i2a( p, f->tag );
		*p++ = TAB;
		memcpy(p, f->val, f->len);
		p += f->len;
#if 0
		e = (s = f->val) + f->len;
		for (;;) {
			while ( s<e && LF != (*p++ = *s++) ) /* tight */
				;
			if ( s==e )
				break;
			if ( opt )
				p[-1] = repl[opt];
			else
				*p++ = TAB;
		}
#endif
		*p++ = LF;
	}
	*p++ = LF;
	return p - buf;
}	/* rSer */


#define REVBIT 0x80000000 /* highest bit on the length */
#define LENMSK 0x7fffffff /* will cut fields longer 2GB */

/*
	basic weak-heap-sort (Dutton 93)
	for faster variants see indexed-relaxed-weak-heap-sort (Edelkamp)
	or, for slightly better avg, quick-weak-heap-sort
*/
void rSort ( Fld *r, VGt *gt )
{
	Fld head, tmp;
	int l = -r->tag -1;
	int i;

	if ( 2 > l )
		return;
	/* fields to sort are r[1]..r[l]
		we use r[0]..r[l-1], since whs puts biggest at 0 in first step
		and sorts the others at 1..l-1
	*/
	head = *r;
	*r = r[l];
	/* weak-heap-forestify
		every node roots a weak heap AND is <= root
		(especially the leftist independent roots are <= root)
	*/
	for ( i=l; --i; ) {
		int g = i; /* granma is 1st left i or root, if i is totally leftist ;) */
		while (!(1&g)) /* we're left child */
			g >>= 1; /* shift granma right :( */
		g >>= 1;
		r[i].len &= LENMSK;
		/* merge: swap if g < i */
		if ( !gt(r+i, r+g) )
			continue;
		tmp = r[i]; r[i] = r[g]; r[g] = tmp; /* swap */
		r[i].len |= REVBIT;
	}

	/* re-weak-heap-forestify */
	for ( i=l; --i>1; ) {
		int j=1, left;
		r[i].len &= LENMSK; /* i's REVBIT no longer needed */
		/* follow path of left childs (independent roots) */
		while ( i > (left = (j<<1) + ((REVBIT & r[j].len)?1:0)) )
			j = left;
		/* put max of them and i (new root) at i */
		do {
			/* merge: swap if i < j */
			/* clear all REVBITs, as gt will probably access .len */
			int rj = REVBIT & r[j].len;
			r[j].len &= LENMSK; /* now i,j both are clean */
			if ( !gt(r+j, r+i) ) {
				r[j].len |= rj;
				continue;
			}
			tmp = r[i]; r[i] = r[j]; r[j] = tmp; /* swap */
			/* toggle original REVBIT at j, which now is at i */
			if ( ! rj )
				r[j].len |= REVBIT;
		} while (j >>= 1);
	}
	r[1].len &= LENMSK; /* 2... have been cleared */
	/* initial root still untouched biggest (never had REVBIT) */
	r[l] = *r;
	*r = head;
}	/* rSort */


void rSortTag ( Fld *r ) /* could use radixsort here ... */
{
	Fld head, tmp;
	int l = -r->tag -1;
	int i;

	if ( 2 > l )
		return;
	head = *r;
	*r = r[l];
	for ( i=l; --i; ) {
		int g = i;
		while (!(1&g))
			g >>= 1;
		g >>= 1;
		r[i].len &= LENMSK;
		if ( r[g].tag >= r[i].tag )
			continue;
		tmp = r[i]; r[i] = r[g]; r[g] = tmp; /* swap */
		r[i].len |= REVBIT;
	}

	for ( i=l; --i>1; ) {
		int j=1, left;
		while ( i > (left = (j<<1) + ((REVBIT & r[j].len)?1:0)) )
			j = left;
		do {
			if ( r[i].tag >= r[j].tag )
				continue;
			tmp = r[i]; r[i] = r[j]; r[j] = tmp; /* swap */
			if ( REVBIT & r[i].len )
				r[j].len &= LENMSK;
			else
				r[j].len |= REVBIT;
		} while (j >>= 1);
	}
	r[l] = *r;
	*r = head;
	while ( l-- ) /* clearing them later is significantly faster */
		(++r)->len &= LENMSK;
}	/* rSortTag */


void rSortVal (Fld *r)
{
	Fld head, tmp;
	int l = -r->tag -1;
	unsigned cl; /* compare length */
	int i, cmp;

	if ( 2 > l )
		return;
	head = *r;
	*r = r[l];
	for ( i=l; --i; ) {
		int g = i;
		while (!(1&g))
			g >>= 1;
		g >>= 1;
		if ( (cl = r[g].len) > r[i].len )
			cl = r[i].len;
		if ( !cl ? !r[i].len
			: (0 < (cmp = memcmp(r[g].val,r[i].val,cl))
				|| (!cmp && r[i].len == cl))
		) /* i <= g: ok */
			continue;
		tmp = r[i]; r[i] = r[g]; r[g] = tmp;
		r[i].len |= REVBIT;
	}

	for ( i=l; --i>1; ) {
		int j=1, left;
		r[i].len &= LENMSK;
		while ( i > (left = (j<<1) + ((REVBIT & r[j].len)?1:0)) )
			j = left;
		do {
			unsigned clj = LENMSK & r[j].len;
			if ( (cl = r[i].len) > clj )
				cl = clj;
			if ( !cl ? !r[j].len
				: (0 < (cmp = memcmp(r[i].val,r[j].val,cl))
					|| (!cmp && clj == cl))
			) /* j <= i: ok */
				continue;
			tmp = r[i]; r[i] = r[j]; r[j] = tmp;
			if ( REVBIT & r[i].len ) { /* toggle j's REVBIT */
				r[i].len &= LENMSK;
				r[j].len &= LENMSK;
			} else
				r[j].len |= REVBIT;
		} while (j >>= 1);
	}
	r[1].len &= LENMSK;
	r[l] = *r;
	*r = head;
}	/* rSortVal */



/* ************************************************************
	list functions
*/

List *lInit ( List *l, const char *fmt, ...  )
{
	memset( l, 0, sizeof(*l) );
	l->blk = &l->bl0;
	l->fld = l->fl0;
	l->fav = DEFFIELDS;
	l->bok = l->buf = l->bl0.byt;
	l->end = l->buf + DEFBLKLEN;
	l->siz = 0;
	if ( fmt ) { /* print header */
		va_list ap;
		va_start( ap, fmt );
		lOut( l, -1, 0, fmt, ap );
		va_end( ap );
	}
	return l;
}	/* lInit */


#define	LBLK_FREE( l ) \
	while ( (l)->blk != &(l)->bl0 ) { \
		LBlk *__nxt = (l)->blk->nxt; \
		mBlkFree( (l)->blk ); \
		(l)->blk = __nxt; \
	}


List *lClr ( List *l )
{
	l->fav -= l->fld->tag;
	l->fld->tag = 0;
	l->fld->val = l->bl0.byt;
	l->fld->len = 0;
	l->bok = l->buf = l->bl0.byt;
	l->end = l->buf + DEFBLKLEN;
	l->siz = 0;
	LBLK_FREE(l);
	return l;
}	/* lClr */


List *lReset ( List *l )
{
	if ( l->fld->val != l->bl0.byt ) { /* header has moved */
		if ( l->fld->len > DEFBLKLEN )
			l->fld->len = DEFBLKLEN; /* truncate header !!! */
		memmove( l->bl0.byt, l->fld->val, l->fld->len );
		l->fld->val = l->bl0.byt;
	}
	l->fav += -1 - l->fld->tag;
	l->fld->tag = -1;
	l->bok = l->buf = l->bl0.byt + l->fld->len;
	l->end = l->buf + DEFBLKLEN;
	l->siz = l->fld->len;
	LBLK_FREE(l);
	return l;
}	/* lReset */


void OPT_STDCALL lFini ( List *l )
{
	LBLK_FREE(l);
	if ( l->fld != l->fl0 )
		mFldFree( l->fld );
} /* lFini */


int OPT_REGPARM lExtend ( List *l, unsigned need, int fields )
{
	LBlk *nju;
	Fld *contig;

	if ( 0 <= fields )
		contig = 0;
	else {
		contig = LLAST(l);
		fields = -fields -1;
		need += contig->len;
	}
	LOG_DBG(LOG_VERBOSE, "lExtend need %u avl %d fields %u fav %d ctg %d",
		need, LAVL(l), fields, l->fav, contig);

	if ( l->fav < (unsigned)fields ) {
		int had = - l->fld->tag;
		Fld *f;

		if ( fields < had>>2 ) /* grow 25% */
			fields = had>>2;
		else if ( fields < DEFFIELDS )
			fields = DEFFIELDS;
		f = mFldAlloc(fields + had);
		memcpy( f, l->fld, had*sizeof(Fld) );
		if ( l->fld != l->fl0 )
			mFldFree( l->fld );
		l->fld = f;
		l->fav = (unsigned)fields;
	}

	if ( LAVL(l) >= need )
		return 1;

	l->siz += l->buf - l->blk->byt; /* add to secondary size */
	nju = mBlkAlloc(need + DEFBLKLEN/2);
	nju->nxt = l->blk;
	l->blk = nju;
	l->buf = nju->byt;
	l->end = l->buf + nju->siz; /* >= requested */
	if ( contig ) {
		memcpy( l->buf, contig->val, contig->len );
		contig->val = l->buf;
		l->buf += contig->len;
	}
	return 1;
}	/* lExtend */

static const char at[] = { TAB, '@' };

int lArgv ( List *l, int tag, int argc, const char **argv )
{
	LNEWF( l, tag, 4000 );
	if ( !argc )
		return 0;
	if ( '-' != **argv ) {
		const char *a = *argv++;
		argc--;
		LAPPS( l, a );
	}
	while ( argc-- ) {
		const char *a = *argv++;
		unsigned len = 1;
		if ( '-' == *a )
			a++;
		else
			len = 2;
		LAPP( l, at, len );
		LAPPS( l, a );
	}
	return LLAST(l)->len;
}	/* lArgv */


List *lVar ( List *l, int argc, const char **argv )
{
	static const char tab = TAB;
	LNEWF( l, -1, 4000 );
	if ( !argc )
		return l;
	if ( '-' != **argv ) {
		const char *a = *argv++;
		argc--;
		LAPPS( l, a );
	}
	while ( argc-- ) {
		const char *a = *argv++;
		if ( '-' == *a ) {
			a++;
			LAPP( l, &tab, 1 );
		} else {
			LNEWF( l, 1, 100 );
		}
		LAPPS( l, a );
	}
	return l;
}	/* lVar */


int lCpy ( List *l, const Fld *src, unsigned need )
{
	int len = RLEN(src);
	const Fld *e = src + len;

	if ( !need )
		need = rSiz( src );
	if ( !lExtend(l, need, len) )
		return 0;
	for ( ;src < e; src++ )
		LADDF( l, src );
	return len;
}	/* lCpy */


int lParse (List *l, const char *p, int s)
{
	int state = LPS_VAL&s, bits = (LPS_NEG|LPS_CR)&s;
	const char *e = p + (LPS_LEN&s), *lf;
	unsigned len;
	Fld *f = LLAST(l);

	for (;;) switch (state) {
	case LPS_SOR:
	case LPS_SOL:
		switch (*p) {
		case CR:
			if (e == p+1) goto done; /* ignore */
			if (LF != p[1]) break; /* weird crap */
			p++;
		case LF: /* got blank line */
			return e - p - 1;
		}
		LNEWF(l, 0, 80);
		if ( (LPS_SOR==state) && (CT_IS(D,*p) || '-' == *p) ) {
			eRr(LOG_WARN, "no message name");
			LNEWF(l, 0, 80);
		}
		f = LLAST(l);
		state = LPS_TAG;
		if ( '-' != *p )
			bits &= ~LPS_NEG;
		else {
			bits |= LPS_NEG;
			if ( e == ++p ) goto done;
		}
	case LPS_TAG:
		while ( CT_IS(D,*p) ) {
			f->tag = f->tag*10 + b36val[(unsigned char)*p];
			if ( e == ++p ) goto done;
		}
		if ( LPS_NEG & bits )
			f->tag = -f->tag;
		LOG_DBG(LOG_DEBUG, "got tag %d", f->tag);
		state = LPS_VAL;
		if ( TAB == *p && e == ++p ) goto done;
	case LPS_VAL:
		lf = memchr(p, LF, e-p);
		if ( !lf )
			len = e-p;
		else if ( (len = lf-p) && CR == lf[-1] ) {
			if ( LPS_CR & bits )
				len--;
			else if ( f == l->fld ) {
				len--;
				bits |= LPS_CR;
				LOG_DBG(LOG_DEBUG, "detected crlf");
			}
		}
		LAPP(l, p, len);
		if ( lf ) {
			state = LPS_SOL;
			/* if ( f == l->fld ) STRIP W\t ? */
			if ( e > (p = lf+1) )
				continue;
		}
		goto done;
	} /* for switch */
done:
	return state|bits;
}	/* lParse */


int lOut ( List *l, int tag, const char *fmt, ... )
{
	va_list ap;
	int indir = !fmt;
	int bav;
	int len=0, i;
	unsigned u;
	char *p, *q=0, c=0;
	const char *tok = 0;
	
	va_start( ap, fmt );
	if ( indir ) {
		va_list ap2;
		fmt = va_arg(ap, const char*);
		ap2 = va_arg(ap, va_list);
		va_end( ap );
		ap = ap2;
	}
	LNEWF( l, tag, 4000 );
	bav = LAVL(l);
	p = l->buf;
	for (;;) {
		if ( !*fmt )
			goto extend;
		if ('%' != *fmt) {
			tok = fmt+1;
			while (*tok && '%'!=*tok)
				tok++;
			if ( bav < (len = tok-fmt) )
				goto extend;
			memcpy(p, fmt, len);
			p += len;
			bav -= len;
			if ( !*(fmt = tok) )
				goto extend;
		}
		/* now '%' == *fmt */
		switch (c = *++fmt) {
		case 0: /* bah -- trailing % */
			goto extend;
		case '0': /* only in %0nx */
			if ( fmt[1] < '1' || '8' < fmt[1] || 'x' != fmt[2] )
				continue;
			c = fmt[1];
			tok = fmt + 2;
		case '2': /* meaning %02x */
		case '4':
		case '8':
			if ( 'x' == fmt[1] ) tok = fmt + 1; /* optional */
			if (0) {
		case 'x':
				c = '4';
		case 'd':
		case 'u':
				tok = fmt;
			}
			if ( bav < (len = 12) )
				goto extend;
			u = va_arg(ap, unsigned);
			switch (c) {
			case 'd': len = i2a(p, (int)u); break;
			case 'u': len = u2a(p, u); break;
			default:
				i = len = c - '0';
				while (i--) {
					p[i] = b36dig[u & 0xf];
					u >>= 4;
				}
			}
			fmt = tok;
			break; /* the numbers */
		case 'c':
			if ( bav < (len = 1) )
				goto extend;
			*p = (char)va_arg(ap, int);
			break;
		default:
		case '%':
			if ( bav < (len = 1) )
				goto extend;
			*p = c;
			break;
		case 'b': /* bytes -- eats 2 args len, q */
			len = 2*va_arg(ap, int);
			q = va_arg(ap, char*);
			if ( bav < len )
				goto extend;
		the_bytes_thing:
			for (i=0; i<len; ) {
				u = (unsigned char)*q++;
				p[i++] = b36dig[u>>4];
				p[i++] = b36dig[u & 0xf];
			}
			break;
		case '.': /* only used as .*s */
			if ( '*' != *++fmt || 's' != *++fmt )
				continue; /* prolly bad ... */
		case '*': /* shorthand */
			len = va_arg(ap, int);
		case 's':
			q = va_arg(ap, char*);
			if ( 's' == c )
				len = strlen(q);
			if ( 0 ) {
				Fld *v;
		case 'v':
				v = va_arg(ap, Fld*);
				len = v->len;
				q = v->val;
			}
			if ( bav < len )
				goto extend;
		the_string_thing: /* for eivind aarset */
			/* since we need to shift args to check len and cannot unshift */
			memcpy(p, q, len);
		}
		p += len;
		bav -= len;
		fmt++;
		continue;
	extend:
		LLAST( l )->len += p - l->buf;
		l->buf = p;
		if ( !*fmt )
			break;
		/* need len */
		lExtend(l, len, -1);
		bav = LAVL(l);
		p = l->buf;
		switch (c) {
		case '.':
		case '*':
		case 's':
			goto the_string_thing;
		case 'b':
			goto the_bytes_thing;
		}
		fmt--; /* simply start over */
	}
	if ( !indir ) /* else it's not ours and caller should va_end */
		va_end( ap );
	return len;
}	/* lOut */
