<?php
/*
	The Malete project - the Z39.2/Z39.50 database framework of OpenIsis.
	Version 0.9.x (patchlevel see file Version)
	Copyright (C) 2001-2004 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	see README for more information
EOH */

// $Id: Db.php,v 1.2 2004/11/02 13:44:24 kripke Exp $


/**
	This class represents a "database".

	@version $Revision: 1.2 $
	@license LGPL
	@package Isis
*/
class Isis_Db {
	/** "field definiton table"
		simple name -> number hash (where name should be an "identifier")
		may contain some additional entries using non-identifier chars ...
	*/
	var $fdt;
	var $name;
	var $srv;

	function Isis_Db ( $fdt = null, $name = null, $server = null )
	{
		$this->fdt  = $fdt;
		$this->name = $name;
		$this->srv  = $server;
	}


	/*
		internal helper to construct and send a request
		set $this->srv->dbg to true to get full listing of what's going on
	*/
	function req ( $type, $arg, $emb = null, $lst = null )
	{
		$req = new Isis_Rec();
		$req->head = $this->name.'.'.$type;
		if ( $arg ) $req->head .= "\t$arg";
		if ( $emb ) foreach ($emb as $r) $req->embed($r);
		if ( $lst ) foreach ($lst as $l) $req->append(0,$l);
		return $this->srv->request( $req );
	}


	/*
		query
		with $expr = null, fetch more results from previous query
		with $recs, fetch an array of records, else of mfns
	*/
	function query ( $expr = null, $recs = true, $head = null )
	{
		if ( $expr && $recs && false === strpos($expr, '?') )
			$expr .= '?'; // force fetch records
		$ret = $this->req('Q', $expr);
		$head = $ret->head;
		return $recs ? $ret->recs($this) : $ret->get(0);
	}

	/*
		read one or an array of mfns
		return one or an array of records
	*/
	function read ( $mfn )
	{
		if ( is_array($mfn) ) {
			$ret = $this->req('R', null, null, $mfn);
			return $ret->recs($this);
		}
		$ret = $this->req('R', null, null, array($mfn));
		$recs = $ret->recs($this);
		return $recs[0];
	}

	/*
		read array of terms
		return one or an array of records
	*/
	function terms ( $from, $to = null )
	{
		if (!is_null($to)) $from .= "\t$to";
		$ret = $this->req('T', $from);
		return $ret->get(0);
	}

	/*
		write one or an array of records
		return an array of mfns written
	*/
	function write ( $rec )
	{
		if ( !is_array($rec) ) $rec = array($rec);
		$ret = $this->req('W', null, $rec);
		return $ret->get(0);
	}

	/*
		unlike the other methods, this expects $req to be a prepared X request
		however, $name.X is prepended
		return res->head, which should be a comment
	*/
	function index ( $req )
	{
		$pfx = $this->name.'.X';
		if ( $req->head )
			$req->head = $pfx."\t".$req->head;
		else
			$req->head = $pfx;
		$res = $this->srv->request( $req );
		return $res->head;
	}

}	// class Isis_Db
?>
