<?php
/*
	The Malete project - the Z39.2/Z39.50 database framework of OpenIsis.
	Version 0.9.x (patchlevel see file Version)
	Copyright (C) 2001-2004 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	see README for more information
EOH */

// $Id: Http.php,v 1.2 2004/11/03 13:56:57 kripke Exp $


/**
	This class represents the connection to a HTTP server.
	This is not for general purpose, but to transfer Isis records over HTTP.

	@version $Revision: 1.2 $
	@license LGPL
	@package Isis
*/
class Isis_Http extends Isis_Server {
	var $head;
	/**
		whether we need PHP's nutty v%d[] on parameters
	*/
	var $nuts = 0;


	/** static
		convert Isis_Rec to urlencoded parameters.
		should be a method of Isis_Rec,
		however, since feeble PHP has to parse all the cruft
		on every single request, better limit the code to where it's needed.
	*/
	function toUrl ( $rec, $nuts = 0 )
	{
		$res = '';
		if ($i = $rec->len())
		for ( $res = '', $t = reset($rec->tag), $v = reset($rec->val); ;
			$t = next($rec->tag), $v = next($rec->val)
		) {
			if ( $nuts )
				// use v42[]: the [] so we can send multiple values,
				// and the v cause numerical named params as []
				// tend to crash PHP; GET always, POST sometimes works ...
				// (it's such a feeble ugly pile of shirt)
				$res .= 'v' . $t . '%5B%5D=' . urlencode($v);
			else
				$res .= $t . '=' . urlencode($v);
			if ( !--$i ) break;
			$res .= '&';
		}
		return $res;
	}


	/** static
		we ARE on the nutty side -- try to turn http request into something usable
		see toUrl nuts above for the format
		UNFORTUNATELY we have no way to tell the original order of parameters.
		hmm ... for get we could resort to parse QUERY_STRING manually
		@param mixed $db_or_fdt a Isis_Db object or just it's fdt array
			to look up names.
	*/
	function fromReq ( $db_or_fdt = null, $server_map = null )
	{
		if ( is_object($db_or_fdt) ) {
			$req = new Isis_Rec( '-db', $db_or_fdt );
			$fdt = $db_or_fdt->fdt;
		} else {
			$req = new Isis_Rec;
			$fdt = $db_or_fdt; // array or null
		}
		if ( $server_map )
			foreach ($server_map as $k => $v)
				if (isset($_SERVER[$k]))
					$req->append( $v, $_SERVER[$k] );
		foreach (array($_GET,$_POST) as $in)
		foreach ($in as $k => $v) {
			if ( $fdt ) {
				if ( !is_int($k = $fdt[$k]) )
					continue;
			} elseif ( is_numeric($k) )
				$k = (int)$k;
			elseif ( sscanf( $k, 'v%d', &$i ) )
				$k = (int)$i;
			else
				continue;
			if (is_string($v))
				$req->append($k, $v);
			else foreach ($v as $w) // Wolfsburg
				$req->append($k, $w);
		}
		return $req;
	}


	/**
		pers 1: use pfsockopen and keep-alive
		pers -1: only keep-alive (for multiple requests within one run)
		pers 0: neither
	*/
	function Isis_Http ( $host, $url, $php = 0, $port = 80, $pers = 0 )
	{
		$this->host = $host;
		$this->port = $port;
		$this->pers = 0 < $pers;
		$this->head = 'POST '.$url." HTTP/1.1\r\n"
			. 'Host: '.$host."\r\n"
			. "User-Agent: Isis\r\n"
			// try keep-alive. fnord does not support that with cgi
			. ($pers ? "Connection: keep-alive\r\n" : '')
			. "Content-Type: application/x-www-form-urlencoded\r\n"
			. 'Content-Length: ';
		$this->nuts = $php;
		$this->open();
	}

	

	function request ( $req )
	{
		if ( !$this->sock && !$this->open() )
			return null;
		$url = Isis_Http::toUrl( $req, $this->nuts );
		fwrite($this->sock, $this->head . strlen($url) ."\r\n\r\n" . $url );
		// as opposed to folklore (http://www.php.net/manual/en/ref.stream.php),
		// there is NO trailing CRLF in 1.1 (RTFM err ... RFC 2068 4.1 Page 30)
		fflush($this->sock);
		$ka = 0; // keepalive -- default for 1.1, explicit for 1.0
		$ok = ($line = fgets($this->sock))
			&& sscanf( $line, "HTTP/1.%d 200 OK", &$ka );
		$chunk = 0;
		while ( ($line = fgets($this->sock)) && trim($line) ) {
			// echo "---$line---";
			if ( !strncasecmp('connection: keep-alive', $line, 22) )
				$ka = $ok;
			elseif ( !strncasecmp('connection: close', $line, 17) )
				$ka = 0;
			elseif ( !strncasecmp('transfer-encoding: chunked', $line, 26) )
				$chunk = 1;
		}
		if ( !$ok )
			$res = null;
		else {
			$txt = '';
			if ($chunk) {
				// RFC 2068 3.6 Page 25, 19.4.6 Page 154
				while ( ($line = fgets($this->sock)) && trim($line)
					&& sscanf( $line, "%x", &$chunk ) && $chunk
				) {
					// error_log( "///chunk size $chunk///\n" );
					$data = fread( $this->sock, $chunk );
					$txt .= $data;
					fread( $this->sock, 2 ); // a CRLF
				}
				while ( ($line = fgets($this->sock)) && trim($line) ) // eat footer
					; // echo "???$line???\n";
			} else
				// HERE we rely on our other end being cooperative,
				// nicely ending response with blank line
				while ( ($line = fgets($this->sock)) && trim($line) ) {
					// echo "+++$line+++";
					$txt .= $line;
				}
			$res = new Isis_Rec;
			$res->parse($txt, ISIS_REC_TEXT);
		}
		if ( !$ka ) {
			fclose( $this->sock );
			$this->sock = 0;
		}
		return $res;
	}

}	// class Isis_Http
?>
