<?php
/*
	The Malete project - the Z39.2/Z39.50 database framework of OpenIsis.
	Version 0.9.x (patchlevel see file Version)
	Copyright (C) 2001-2004 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	see README for more information
EOH */

// $Id: Rec.php,v 1.2 2004/11/03 13:57:15 kripke Exp $

/** FIELD mode replaces newlines with tabs.
	On deserializing, these tabs are not converted back to newline.
	Do not use if you need to retain newline information.
*/
define( 'ISIS_REC_FIELD', "\t" );
/** TEXT mode replaces newlines with vertical tabs.
	Vertical tabs are converted back to newlines only when explicitly
	deserializing in TEXT mode, since it's not transparent to binary data.
*/
define( 'ISIS_REC_TEXT', "\013" );
/**
	the basic ISIS formatting mode.
	"Data Mode" MDL doesn't make a significant difference,
	and uppercase can better be achieved using strtoupper
	(given the right locale ...).
*/
define( 'ISIS_REC_MHL', 'MHL' );
// sorry, no other way to save this ...
$GLOBALS['ISIS_REC_MHL_PAT'] = array(
	'/(<[^=>]*)=[^>]+>/', # dump <a=b> substitutions
	'/></', # replace >< pairs
	'/[><]/', # nuke other ><
	'/^\^./', # kill initial subfield spec
	'/\^a/', # ^a -> ;
	'/\^[b-i]/', # ^[b-i] -> ,
	'/\^./' # others -> .
);
$GLOBALS['ISIS_REC_MHL_REPL'] = array( 
	'\1>', '; ', '', '', '; ', ', ', '. '
);

/**
	An ISIS(/IIF/Z39.2/ISO2709)-style record in pure PHP.

	This is only loosely connected to an Isis Database,
	most functions can be used without having a DB.

	Note: Most getter methods work regardless of the type of tags.
	However, the setters are strongly biased towards numeric tags.

	@version $Revision: 1.2 $
	@license LGPL
	@package Isis
*/
class Isis_Rec {
	/**
		The associated database, if any. Better be an Isis_Db.
		@var object $db
	*/
	var $db = 0;
	/**
		The record's MFN ("master file number", a.k.a. rowid).
		A value of 0 means the record did not yet have a number in it's db,
		a write will be handled as insert.
		@var int $mfn
	*/
	var $mfn = 0;
	/**
		The record's header.
		For data records, this is mfn[@pos][\tleader],
		where leader can any be any MARC leader data.
	*/
	var $head = '';
	/**
		Array of tags for the record.  Keys and values better be integers
		(others might be ignored or yield unexpected results in some contexts).
		The count of the tag array is taken as length of the record.
		The tag array may have holes (unassigned ints) if unset is used.
		@var array $tag
	*/
	var $tag;
	/**
		Array of values for the record. Values better be strings
		(i.e. will be forced to strings by some operations).
		@var array $tag
	*/
	var $val;
	/**
		last tag accessed using function v
	*/
	var $v = null;



//	////////////////////////////////////////////////////////////
//	static
//

	/** static function to format a value.
		maybe used as Isis_Rec::fmt.
		@param string $val a value to be formatted
		@param string $fmt a format specification, defaults to 'MHL'
		- if $fmt is null,
			the value is returned unchanged
		- if $fmt starts with '&' or '%',
			that character is stripped and htmlspecialchars or urlencode, resp.,
			is applied to the (each) value as last step
		-	if $fmt starts with 'MHL' (the constant ISIS_REC_MHL),
			that is stripped and the classical MHL ISIS formatting applied to values
			(before & or % mangling, rarely needed with subfields)
		- if $fmt is (now) empty,
			the complete value is used
		-	else we're going for subfields:
		- if $fmt starts with a hat,
			the hat is stripped an used as subfield delimiter (TAB otherwise)
		- if $fmt matches /^([^(]*)\((\d*)(\.\.(\d*))?/,
			the part from the first '(' on is stripped as occurence selector
			(note that an optional closing ')' and additional chars are ignored)
		-	the (remaining) characters in $fmt are subfield names,
			'*' selects any subfield
				(including the initial unnamed, even if it's empty !),
			'' any without stripping subfield names
		-	if there is a occurence or range selected,
			for every character in the $fmt the specified occurences are used
			(counted from 0). If either bound is empty, 0 is used.
			If a range is specified (.. given), an upper bound of 0 means up to end.
			By default, only the first occurence (0) is used.
		@return if there is only a single character (remaining) in $fmt,
			and only a single occurence selectedi (no ..), a string is returned.
			Else you've been asking for an array,
			and so an array is returned even if it contains only a single value.
			If more than one subfield name was specified or the '*',
			the names are used ('' for the initial).
			If a range is selected, the index is used (in addition).
			Fields are added by first looping over subfield names,
			then occurences. PHP may or may not loop the array in that order.
		Example:
			'^ab' gives keys 'a' and 'b', 'b(1..' gives keys '1','2'...m
			'^cab(..', gives 'c0', 'c1', ... 'a0', ...
	*/
	function fmt ( $val, $fmt = ISIS_REC_MHL )
	{
		global $ISIS_REC_MHL_PAT, $ISIS_REC_MHL_REPL; 
		if ( null == $fmt )
			return $val;
		if ( '&' != $fmt{0} && '%' != $fmt{0} )
			$mode = '';
		else {
			$mode = '&' == $fmt{0} ? 'htmlspecialchars' : 'urlencode';
			$fmt = substr($fmt,1);
		}
		if ($mhl = ('M' == $fmt{0} && ISIS_REC_MHL == substr($fmt,0,3)) )
			$fmt = substr($fmt,3);
		if ( '' == $fmt ) {
			if ( $mhl )
				$val = preg_replace($ISIS_REC_MHL_PAT, $ISIS_REC_MHL_REPL, $val);
			return $mode ? $mode($val) : $val;
		}
		if ( '^' != $fmt{0} )
			$sep = "\t";
		else {
			$sep = '^';
			$fmt = substr($fmt,1);
		}
		$sub = explode($sep,$val); // this really is not for performance ...
		$n = count($sub);
		if ( false !== strpos($fmt,'(') // quick check
			&& preg_match( '/^([^(]*)\((\d*)(\.\.(\d*))?/', $fmt, $m )
			// actually matches everything with a ( in it
		) {
			$fmt = $m[1]; // initial part
			if (!($fst = $m[2]) )
				$fst = 0;
			if (!($range = $m[3]))
				$lst = $fst;
			elseif (!($lst = $m[4]))
				$lst = $n;
		} else
			$fst = $range = 0;
		$strip = $l = strlen($fmt);
		if (!$strip) { // plain mode
			$fmt = '*'; $l = 1;
		}
		if ( !($single = 1 == $l && !$range) )
			$ret = array();
// echo "sub '$fmt'($fst..$lst) range $range single $single strip $strip n $n\n";
		for ($i = 0; $i < $l; $i++ ) {
			$s = $fmt{$i};
			$sel = '*' == $s ? 0 : 1;
			$o = -1;
			for ($j = $sel; $j < $n; $j++ ) {
				// echo "test '$s'($fst..$lst) $o\n";
				if ( ($sel && $s != $sub[$j]{0}) || ++$o < $fst )
					continue;
				if ( $lst < $o )
					break;
				// ok, it's in range
				$v = ($strip && $j) ? substr($sub[$j],1) : $sub[$j];
				if ( $mhl )
			 		$v = preg_replace($ISIS_REC_MHL_PAT, $ISIS_REC_MHL_REPL, $v);
				if ( $mode )
					$v = $mode($v);
				if ( $single )
					return $v;
				$key = $sel ? (1 == $l ? '' : $s)
					: $strip && $j ? $sub[$j]{0} : '';
				if ($range) $key .= $o;
				$ret[$key] = $v;
				if ( !$range )
					break;
			}
		}
		return $ret;
	}


//	////////////////////////////////////////////////////////////
//	ctor
//

	/**
		create an ISIS record.
		@return object Isis_Rec a new ISIS record
	*/
	function Isis_Rec ()
	{
		$this->tag = array();
		$this->val = array();
		if ( func_num_args() )
			$this->add( func_get_args() );
	}



//	////////////////////////////////////////////////////////////
//	getters
//

	/**
		@return the number of fields
	*/
	function len ()
	{
		return count($this->tag);
	}


	/**
		try to look up non-numeric tags in the fdt
	*/
	function fdt ( &$tag )
	{
		if ( !is_int($tag) ) {
			if (is_numeric($tag))
				$tag = (int)$tag;
			else if ($this->db && $this->db->fdt && is_int($this->db->fdt[$tag]))
				$tag = $this->db->fdt[$tag];
		}
	}


	/**
		reset tag and val array pointers
	*/
	function res ()
	{
		reset($this->tag);
		reset($this->val);
	}


	/** get all values for tag as array
	*/
	function get ( $tag )
	{
		$ret = array();
		$this->fdt($tag);
		reset($this->tag);
		while (list($p,$t) = each($this->tag))
			if ( $tag == $t )
				$ret[] = $this->val[$p];
		return $ret;
	}


	/** v is for value
		-- get the value of the next occurence of tag in the record.
		The position is reset by res or when using v with a different tag.
		@param mixed $fmt format to apply
		-	if $fmt is null (or the value is null),
		the value is returned unmodified
		-	if $fmt is a string,
		the value is formatted by Isis_Rec::fmt (returns string or array).
		- other values of $fmt are reserved for future extensions
		and currently return null.
	*/
	function v ( $tag = null, $fmt = null )
	{
		if ( !is_int($tag) )
			$this->fdt($tag);
		if ( $this->v != $tag ) {
			reset($this->tag); // no each on val needed
			$this->v = $tag;
		}
		$v = null;
		if ( is_null($tag) ) {
			if (list($p,$t) = each($this->tag))
				$v = $this->val[$p];
		} else
			while (list($p,$t) = each($this->tag))
				if ( $tag == $t ) {
					$v = $this->val[$p];
					break;
				}
		if (is_null($v) || is_null($fmt))
			return $v;
		if ( is_string($fmt) )
			return Isis_Rec::fmt($v,$fmt);
		return null;
	}

	/**
		return an array of subrecords
	*/
	function recs ( $db = null )
	{
		$ret = array();
		$n = 0;
		reset($this->tag);
		while (list($p,$t) = each($this->tag))
			if ( 0 > $t ) {
				$r = &$ret[$n++];
				$r = new Isis_Rec();
				$r->db = $db;
				$r->head = $this->val[$p];
				for ( $i=-(int)$t-1; $i-- && list($p,$t) = each($this->tag); ) {
					$r->tag[] = $t;
					$r->val[] = $this->val[$p];
				}
			}
		return $ret;
	}


	/** same as v($tag,'&'.$fmt)
	*/
	function h ( $tag = null, $fmt = '' )
	{
		return $this->v($tag,'&'.$fmt);
	}


	/** same as v($tag,'&MHL')
	*/
	function mhl ( $tag = null )
	{
		return $this->v($tag,'&MHL');
	}



//	////////////////////////////////////////////////////////////
//	setters
//

	/**
		append a new field (tag-value-pair) to the end of the record.
		@param int $tag tag to use in the field. fdt is applied
		@param string $val the new fields value.
			The string type is not enforced here.
		@return the new value
	*/
	function append ( $tag, $val )
	{
		if ( !is_int($tag) )
			$this->fdt($tag);
		// echo "0\tappending $tag ",gettype($val),"\n";
		if ( is_string($val) || is_numeric($val) ) {
			$this->tag[] = $tag;
			$this->val[] = $val;
		} elseif ( is_array($val) ) {
			foreach ($val as $v)
				$this->append($tag, $v);
		} elseif ( is_object($val) )
			$this->embed($val);
		return $val;
	}

	/**
		add an array to the record.
		@param mixed $argv an array either as a single parameter
			or as a variable number of arguments.
		The array is processed as follows:
		-	if an item is an int, it is appended, with the following item as value.
		- if an item is an array, add is called recursively on this array
		- if an item is '-db' or '-mfn', the corresponding properties are set
		- if we have an fdt which maps the item to an int,
			the following item as value is added with the tag given by the fdt.
		- if an item is ISIS_REC_TEXT, the following item is parsed in text mode.
		- else the item is parsed in standard mode.
	*/
	function add ( $argv )
	{
		$added = 0;
		$fdt = $this->db ? $this->db->fdt : null;
		if (1 < func_num_args() || !is_array($argv))
			$argv = func_get_args();
		// standardized to having an array
		for ( $i = reset($argv); $i || !is_bool($i); $i = next($argv) ) {
			if ( is_int($i) ) {
				if ( !is_null($this->append($i, next($argv))) )
					$added++;
			} elseif ( is_array($i) )
				$added += $this->add($i);
			elseif ( '-mfn' == $i )
				$this->mfn = next($argv);
			elseif ( '-db' == $i ) {
				// they don't grok $fdt = ($this->db = next($argv))->fdt;
				$this->db = next($argv);
				$fdt = $this->db->fdt;
			} elseif ($fdt && is_int($fdt[$i])) {
				if ( !is_null($this->append($fdt[$i], next($argv))) )
					$added++;
			} elseif ( ISIS_REC_TEXT == $i )
				$added += $this->parse( next($argv), ISIS_REC_TEXT );
			else
				$added += $this->parse( $i );
		}
	}


	function addglobals ()
	{
		if ( $this->db && $this->db->fdt )
			foreach ($this->db->fdt as $name => $tag)
				$this->append($tag, $GLOBALS[$name]);
	}

	/** recompact after unsetting
	*/
	function pack ()
	{
		$this->tag = array_values( $this->tag );
		$this->val = array_values( $this->val );
	}


	/** remove a field at given pos, poking a hole in the field list.
	*/
	function rm ( $pos, $pack = FALSE )
	{
		unset( $this->tag[$pos] );
		unset( $this->val[$pos] );
		if ( $pack )
			$this->pack();
	}


	/** remove all fields or all with a given tag.
	*/
	function del ( $tag = null, $pack = FALSE )
	{
		if ( is_null($tag) ) {
			$this->tag = array();
			$this->val = array();
		} else {
			if ( !is_int($tag) )
				$this->fdt($tag);
			foreach ($this->tag as $p => $t) 
				if ($tag == $t)
					$this->rm($p);
			if ( $pack )
				$this->pack();
		}
	}


	/** set fields with tag to values.
		set( 42, 'foo', 'bar', 'baz' ) will change
		the first three occurences of 42 to 'foo', 'bar' and 'baz', resp.
		-	if there are less than three occurences,
		the remaining values are appended
		-	if there are more than three occurences,
		the remaining occurences are deleted
		- if a value is the integer 0,
		processing stops (i.e. remaining occurences are left unchanged)
		- if a value is a positive integer n,
		processing skips n occurences (letting them unchanged)
		- if a value is an array,
		it's elements are used (it is flattened out non-recursively)
		@param mixed $tag tag by int or name
		@param mixed values... variable number of values
	*/
	function set ( $tag )
	{
		if ( !is_int($tag) )
			$this->fdt($tag);
		$argv = func_get_args();
		reset($argv);
		each($argv); // eat $tag
		reset($this->tag);
		$ary = null;
		$more = 1;
		for (;;) {
			if ( $ary ) {
				if ( !(list($k,$arg) = each($ary)) ) {
					$ary = null;
					continue;
				}
			} else {
				if ( !(list($k,$arg) = each($argv)) )
					break;
				if ( is_array($arg) ) {
					$ary = $arg;
					reset($ary);
					continue;
				}
			}
			// echo "setting '$arg'\n";
			if (is_int($arg)) {
				if (!$arg)
					return;
				while ($more && 0 < $arg--)
					while (($more = (list($k,$v) = each($this->tag))) && $tag != $v)
						;
				continue;
			}
			// now arg is a value to set/add
			if ($more) {
				while (($more = (list($k,$v) = each($this->tag))) && $tag != $v)
					;
				if ($more) {
					$this->val[$k] = $arg;
					continue;
				}
			}
			$this->append($tag,$arg);
		}
		if ( $more )
			while (list($k,$v) = each($this->tag))
				if ($tag == $v)
					$this->rm($k);
	}


	/* transparently embed a record.
	*/
	function embed ( $that )
	{
		$this->append( -($i = $that->len())-1, $that->head );
		for ( $t = reset($that->tag), $v = reset($that->val);
			$i--;
			$t = next($that->tag), $v = next($that->val)
		) {
			$this->tag[] = $t;
			$this->val[] = $v;
		}
	}


//	////////////////////////////////////////////////////////////
//	conversion to/from other representations
//

	/**
		serialize record to a string.
		After each field, including the last one, a newline is added.
		@param string $mode replacement value for newlines.
			suggested is one of the predefined constants.
			defaults to ISIS_REC_TEXT.
		@return the string representation of the record
	*/
	function toString ( $mode = ISIS_REC_TEXT )
	{
		$s = '';
		if ( strlen($this->head) ) {
			if ( '0'<=$this->head{0} && $this->head{0}<='9' )
				$s .= "W\t";
			$s .= $this->head."\n";
		}
		// why in heaven is next a pre-increment ???
		$t = reset($this->tag);
		$v = reset($this->val);
		for ( $i = count($this->tag);
			$i--;
			$t = next($this->tag), $v = next($this->val)
		)
			$s .= $t . "\t" . str_replace("\n",$mode,$v) . "\n"; 
		return $s;
	}


	/** parse text as record fields to add.
		For each non-empty line, initial digits are used as tag (empty == 0),
		an optional following tab is skipped, and the rest used as value,
		after replacing $repl, if given, with newlines.
		If the line starts with a tab and the record is not empty,
		a newline and the value are appended to the last field,
		else a new field is appended to the record.
		@param string $repl string to be converted back to newlines.
			use ISIS_REC_TEXT, if you know text is from toString(ISIS_REC_TEXT)
		@return number of fields added
	*/
	function parse ( $text, $repl = null )
	{
		// need compact array in order to reliably know last index
		$conv = 0;
		$lines = explode("\n",$text);
		if ( $lines && strlen($lines[0]) ) {
			$line = $lines[0];
			if ( $line{0}<'0' || '9'<$line{0} ) {
				$this->head = $line;
				array_shift($lines);
			}
		}
		foreach ($lines as $line) {
			if ('' == $line) // blank line or trailing newline
				continue;
			$dig = strspn($line,'0123456789-');
			$t = $dig ? (int)substr($line,0,$dig) : 0;
			$v = substr( $line, $dig + ("\t" == $line{$dig} ? 1 : 0) ); 
			if ( $repl )
				$v = str_replace($repl,"\n",$v);
			$this->tag[] = $t;
			$this->val[] = $v;
			$conv++;
		}
		return $conv;
	} // parse


	/**
		return the "data fork" of this record by mapping a function
		to tags and values in parallel.
		@param function $func defaults to null,
			resulting in an array of fields, each an array [0] => $tag, [1] => $val.
		@return a new array as of array_map
	*/
	function map ($func = null)
	{
		return array_map($func,$this->tag,$this->val);
	}
}	// class Isis_Rec
?>
