<?php
/*
	The Malete project - the Z39.2/Z39.50 database framework of OpenIsis.
	Version 0.9.x (patchlevel see file Version)
	Copyright (C) 2001-2004 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	see README for more information
EOH */

// $Id: Server.php,v 1.4 2004/11/12 09:34:55 kripke Exp $


/**
	This class represents the connection to an Isis server.

	In general, a server is any object having a request function,
	accepting a single Isis_Rec parameter and returning an Isis_Rec.
	
	This implementation is based on a TCP or UNIX socket.

	@version $Revision: 1.4 $
	@license LGPL
	@package Isis
*/
class Isis_Server {
	var $host;
	var $port;
	var $sock;
	var $pers;
	var $dbg = false;

	function Isis_Server ($host = null, $port = 2042, $pers = 0)
	{
		if (!$host && !($host = $_ENV['ISIS_SERVER'])) $host = 'localhost';
		$this->host = $host;
		$this->port = $port;
		$this->pers = $pers;
		$this->open();
	}

	function open ()
	{
		// pfsockopen will reuse an existing socket on same host:port
		$this->sock = $this->pers
			? pfsockopen( $this->host, $this->port /*, $errno, $errstr*/ )
			: fsockopen( $this->host, $this->port /*, $errno, $errstr*/ );
		return $this->sock;
	}


	function request ( $req )
	{
		if ( !$this->sock && !$this->open() )
			return null;
		if ($this->dbg) error_log("SEND\n".$req->toString(ISIS_REC_TEXT));
		fwrite($this->sock, $req->toString(ISIS_REC_TEXT) . "\n");
		fflush($this->sock);
		$txt = '';
		while ( ($line = fgets($this->sock)) && "\n" != $line ) {
			if ($this->dbg) error_log("RETR ".$line );
			$txt .= $line;
		}
		$res = new Isis_Rec;
		$res->parse($txt, ISIS_REC_TEXT);
		if ($this->dbg) error_log("GOT ".$res->toString() );
		return $res;
	}

}	// class Isis_Server
?>
