<?php // vim:syntax=php
/*
	The Malete project - the Z39.2/Z39.50 database framework of OpenIsis.
	Version 0.9.x (patchlevel see file Version)
	Copyright (C) 2001-2004 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	see README for more information
EOH */

// $Id: demo.php,v 1.3 2004/11/02 13:44:42 kripke Exp $
// demo for the Isis package

require 'Isis.php'; // use require 'Isis/Rec.php' if you need only this
?><html><head>
	<title>Demo for package Isis</title>
</head><body>
<?php
	
	// get request parameters a and b
	// (as well as any plain numeric and v%d style
	$param = Isis_Http::fromReq( array(
		'a' => 22, 'b' => 42
	) );

	// create a db with field list ("fdt")
	$fdt = array(
		'title' => 24,
		'author' => 70,
		'keywords' => 69
	);
	$db = new Isis_Db($fdt, 'test');

	// a special value using lotsa subfields
	$subs = 'initial	aParis	bUnesco	b<test=foo>	c-1965';
	// create record using various add parameters
	$r = new Isis_Rec(
		// set the db
		'-db', $db,
		// first some lines from CDS, some using field names, some plain int tags
		'keywords', 'Paper on: <plant physiology><moisture><temperature><wind><measurement and instruments><ecosystems>',
		'author', 'Magalhaes, A.C.',
		24, '<The> Controlled climate in the plant chamber',
		76, 'Les Politiques de la communication en Yougoslavie	zfre',
		'author', 'Franco, C.M.',
		26, $subs,
		// a field to test del
		77, 'ave Caesar',
		// a field using tab as subfield separator
		42, "foo\tbar\tbaz",
		// a field containing newline
		99, "two\nlines",
		// a serialized record (as of toString) as parameter
		"70\tyet another author\n99two more\n99lines\na 0 field\n42\tthe\tanswer"
	);

	// dump the stuff
?>
<h2>dump of record</h2>
	have <?=$r->len()?> fields
<pre>--
<?=htmlspecialchars($r->toString())?>--</pre>

<h2>examples of field selector</h2>
<?php

	// more dump -- don't
	// print_r($r->tag); // print_r corrupts data with 4.3.1
	// print_r($r->map());

	$r->del( 77 ); // ... morituri te salutant

	// various examples how to access fields
	foreach ($db->fdt as $name => $tag)
		while ($v = $r->mhl($tag))
			echo "$name = $v<br/>\n";
		
?>

<!-- single select -->
v26^a = <?=$r->h(26,'a')?><br>
<?php
	// can't use multiple of this on same tag w/o resetting,
	// since the magic loop would advance to the next occurence of tag
	$r->res();
	// more convenient to get them all at once
	// the same as Isis_rec::fmt($r->v(26),'&abc'),
	foreach ($r->h(26,'abc') as $k => $v)
		echo "v26^$k = $v<br>\n";
?>

<!-- select by name -->
v69 = <?=$r->mhl('keywords')?><br>
v76 = <?=$r->mhl(76)?><br>
<!-- use tabbed subfield w/o marker -->
v42(1) = <?=$r->h(42,'(1)')?>

<?php
	// more formatting tests on subs
	echo "<h2>formats returning arrays on '$subs'</h2>\n";
	foreach (array(
		'&cab(..',
		'&*(..',
		'&(..',
		'&bca(1..',
		'&MHLcab(..'
	) as $fmt) {
		echo "\n\n<h3>Format &quot;$fmt&quot;</h3>\n";
		foreach (Isis_rec::fmt($subs,$fmt) as $k => $v)
			echo "v26^$k = $v<br/>\n";
	}
?>

<h2>embedding and TEXT mode</h2>
<?php
	$q = new Isis_Rec( 77, 'sunset strip' );
	// embed r into q
	$q->embed( $r );
	$s = $q->toString(ISIS_REC_TEXT);
	echo "<h3>record embedded</h3>\n<pre>---\n$s---\n</pre>\n";
	$q->del();
	// restore from the string
	$q->parse( $s, ISIS_REC_TEXT );
	// extract the original r
	$recs = $q->recs();
	$r = $recs[0];
	$r->db = $db;
	echo "<h3>record restored from field 22</h3>\n",
		"<pre>---\n", $r->toString(), "---\n</pre>\n";
?>

<h2>set operator</h2>
<?php
	// change titles, adding one
	$r->set( 'title', 'new title', 'second new title' );
	// change 99s, removing one
	$r->set( 99, 'now a oneliner' );
	// change author[1] only, also test array parameter
	$r->set( 'author', array(1, 'Blanco', 0) );
	echo "<pre>---\n", $r->toString(), "---\n</pre>\n";
?>

<h2>server</h2>
<?php
	$db = new Isis_Db($fdt, 'test', new Isis_Server());
	if ( !$db->srv->sock )
		echo "could not contact server\n";
	else {
?>
	<h3>terms beginning with a</h3>
<?php
	$terms = $db->terms('a');
	echo "got ",count($terms), "terms for 'a'</br>\n";
	foreach ($terms as $t) {
		list($cnt, $term) = explode("\t", $t);
		echo "'$term'($cnt) ";
	}
	echo "</br>\n";
?>
	<h3>query reading records</h3>
<?php
	$recs = $db->query('plant water');
	echo "got ",count($recs), "records for 'plant water'</br>\n";
	foreach ($recs as $r)
		echo "<pre>---\n", $r->toString(), "---\n</pre>\n";
?>
	<h3>query reading mfns</h3>
<?php
	$mfns = $db->query('plant + water + devel$', false);
	do {
		echo "got ",count($mfns), " mfns: ", join(',',$mfns),"</br>\n";
	}	while ($mfns = $db->query(null,false));
?>
	<h3>reading 42, 43</h3>
<?php
	$recs = $db->read( array(42,43) );
	foreach ($recs as $r)
		echo "<pre>---\n", $r->toString(), "---\n</pre>\n";
?>
	<h3>reading 42</h3>
<?php
	$r = $db->read(42);
	echo "<pre>---\n", $r->toString(), "---\n</pre>\n";
?>
	<h3>writing 42</h3>
<?php
	$r->append('author', 'one more author');
	echo "<pre>---\n", $r->toString(), "---\n</pre>\n";
	$mfns = $db->write($r);
	echo "wrote ",count($mfns), " mfns: ", join(',',$mfns),"</br>\n";
?>
	<h3>writing 42 as new record</h3>
<?php
	$r->head = '';
	$mfns = $db->write($r);
	echo "wrote ",count($mfns), " mfns: ", join(',',$mfns),"</br>\n";
?>
	<h3>indexing author fields as 70 in split mode</h3>
<?php
	$idx = new Isis_Rec();
	$idx->head = 's';
	$idx->set( 70, $r->get('author') );
	echo "<pre>---\n", $idx->toString(), "---\n</pre>\n";
	$res = $db->index($idx);
	echo "got $res</br>\n";
?>
	<h3>query one near author</h3>
<?php
	$mfns = $db->query('one .. author', false);
	echo "got ",count($mfns), " mfns: ", join(',',$mfns),"</br>\n";
?>
<?php
	} // end else could contact server
?>
</body></html>
